# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

from hiq.utils import ts_to_dt, get_env_bool, lmk_data_handler
from hiq.constants import *
import os
import time
from multiprocessing import Process, Queue, Lock

import signal
import functools


# https://www.cloudcity.io/blog/2019/02/27/things-i-wish-they-told-me-about-multiprocessing-in-python/
class SignalObject:
    MAX_TERMINATE_CALLED = 3

    def __init__(self, shutdown_event):
        self.terminate_called = 0
        self.shutdown_event = shutdown_event


def default_signal_handler(
    signal_object, exception_class, signal_num, current_stack_frame
):
    signal_object.terminate_called += 1
    signal_object.shutdown_event.set()
    if signal_object.terminate_called == signal_object.MAX_TERMINATE_CALLED:
        raise exception_class()


def init_signal(signal_num, signal_object, exception_class, handler):
    handler = functools.partial(handler, signal_object, exception_class)
    signal.signal(signal_num, handler)
    signal.siginterrupt(signal_num, False)


"""
def init_signals(shutdown_event, int_handler, term_handler):
    signal_object = SignalObject(shutdown_event)
    init_signal(signal.SIGINT, signal_object, KeyboardInterrupt, int_handler)
    init_signal(signal.SIGTERM, signal_object, TerminateInterrupt, term_handler)
    return signal_object
"""


class LogMonkeyKing(object):
    """A multi-process Log Monkey"""

    @staticmethod
    def consumer_func(log_queue, lock, lmk_path, lmk_handler, logger):
        if os.cpu_count() >= 2 and os.uname().sysname=='Linux':
            affinity_list = list(os.sched_getaffinity(0))
            os.sched_setaffinity(0, set(affinity_list[len(affinity_list) // 2 :]))
        # Synchronize access to the console
        pid = os.getpid()
        # with lock:
        #    print("process 🐒 {}".format(pid))
        if lmk_path:
            if hasattr(logger, "add"):
                logger.add(lmk_path, rotation="500 MB")
        while True:
            # If the log_queue is empty, log_queue.get() will block until the log_queue has data
            data = log_queue.get()
            if data:
                if lmk_handler:
                    data = lmk_handler(data=data, pid=pid)
                else:
                    data = lmk_data_handler(data=data, pid=pid)
                with lock:
                    if logger:
                        logger.info(data)
                    else:
                        print(data)

    def __init__(sf, lmk_path=None, lmk_handler=None, lmk_logger=None, *args, **kwargs):
        # to use a cache ttl
        sf.summon_log_monkey_king = get_env_bool("LMK")
        if not sf.summon_log_monkey_king:
            sf.queue_lmk = sf.consumer = None
            return

        sf.queue_lmk = Queue()
        sf.lock = Lock()
        sf.consumer = Process(
            target=LogMonkeyKing.consumer_func,
            args=(sf.queue_lmk, sf.lock, lmk_path, lmk_handler, lmk_logger),
        )
        # This is critical! The consumer function has an infinite loop
        # Which means it will never exit unless we set daemon to true
        sf.consumer.daemon = True
        sf.consumer.start()

    def __del__(sf):
        sleep_count = 0
        if hasattr(sf, 'queue_lmk'):
            while sf.queue_lmk and sf.queue_lmk.qsize() and sleep_count < 1:
                time.sleep(2)
                sleep_count += 1
        else:
            # TODO
            pass
        if hasattr(sf, 'consumer'):
            if sf.consumer:
                print(sf.consumer)
                sf.consumer.join()
        else:
            # TODO
            pass

    # def tree_send_log_to_monkey(sf, **kwargs):
    #    sf.queue_lmk.put(kwargs)

    @staticmethod
    def __log(id_, extra, key, name, value):
        if len(name) >= 2 and name[:2] == "__":
            name = name[2:]
        _extra = dict(extra)
        monkey = "🐵" if key == EXTRA_START_TIME_KEY else "🙈"
        if key in _extra:
            del _extra[key]
            if not _extra:
                _extra = ""
            # metric here is actually the tree id
            print(
                f"{ts_to_dt(extra[key])} - [{id_}] [🆔 {os.getpid()}] {monkey} [{name}] {value} {_extra}"
            )
        else:
            if not _extra:
                _extra = ""
            if "time" in id_:
                value = ""
            print(
                f"{ts_to_dt(time.time())} - [{id_}] [🆔 {os.getpid()}] {monkey} [{name}] {value} {_extra}"
            )

    @staticmethod
    def log(id_: str, name: str, value, extra: dict, is_start=True, to_stdio=True):
        """log like a normal logger but with monkeyking

        Args:
            name: node name
            value (float or int): node value, it the value itself is time, there will be no time information in extra
            extra: extra information to log
        """
        if to_stdio:
            if is_start:
                LogMonkeyKing.__log(id_, extra, EXTRA_START_TIME_KEY, name, value)
            else:
                LogMonkeyKing.__log(id_, extra, EXTRA_END_TIME_KEY, name, value)
        else:
            pass


if __name__ == "__main__":
    import pickle

    lmk = LogMonkeyKing()
    print(lmk)
    d = pickle.dumps(lmk)
    lmk2 = pickle.loads(d)
    d2 = pickle.dumps(lmk2)
    print(d == d2)



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import ast
import base64

import itree
from hiq.node import Node
from hiq.constants import *
from hiq.utils import ts_pair_to_dt, memoize
from typing import Union


def __create_tmp_node():
    return Node("", 0, 0)


def _d(n: Node, s: str = ""):
    """serialize a tree(the node and its children) to string
    Since A node is well-defined, we can use format like `{name,start,end,duration}` to serialize it.
    """
    if n:
        extra_ = ""
        if n.extra:
            extra_ = dict_to_b64_str(n.extra)
        s += '{"' + str(n.name) + '":' + str(n.start) + "," + str(n.end) + "," + extra_
        for i in n.nodes:
            s = _d(i, s)
        s += "}"
    return s


def is_str_dict(s):
    try:
        r = ast.literal_eval(s)
        if isinstance(r, dict):
            return True
    except:
        return False


def str_to_number(x) -> Union[int, float]:
    x = float(x)
    if x.is_integer():
        return int(x)
    else:
        return x


def __parse_node(s):
    _start, _end, _extra = s.split(",")

    if _extra and _extra != "inf":
        _extra = b64_str_to_dict(_extra)
    # _extra should be a dictionary
    if _extra and isinstance(_extra, str) and not is_str_dict(_extra):
        _extra = {}
    _sta = str_to_number(_start)
    _end = str_to_number(_end)
    if _extra == "":
        _extra = {}
    return _sta, _end, _extra


def _c(bs: str) -> Node:
    """de-serialize a string to a node"""
    if not bs:
        return None
    stk_ = [itree.create_virtual_node()]
    s = ""
    for ch in bs:
        if ch == "{":
            if len(stk_) == 1:
                stk_.append(__create_tmp_node())
                stk_[0].append(stk_[1])
            else:
                if s:
                    kv = s.split(":")
                    stk_[-1].name = kv[0][1:-1]
                    stk_[-1].start, stk_[-1].end, stk_[-1].extra = __parse_node(kv[1])
                    s = ""

                t = __create_tmp_node()
                stk_[-1].append(t)
                stk_.append(t)
        elif ch == "}":
            if s:
                kv = s.split(":")
                stk_[-1].name = kv[0][1:-1]
                stk_[-1].start, stk_[-1].end, stk_[-1].extra = __parse_node(kv[1])
                s = ""
            if len(stk_) > 1:
                stk_.pop()
        else:
            s += ch
    assert len(stk_) == 1, "bad tree data"
    n = itree._itree.consolidate(stk_[0])
    return n


def __peek_tree(node, debug=False):
    """look 2 levels deep to find the meta information of the tree

    Args:
        node (Node): root node of a tree
    Returns:
        wide_output, value_column_len
    """
    wide_output = False
    if node.extra and "start" in node.extra or "end" in node.extra:
        wide_output = True
    if not wide_output and len(node.nodes) > 0:
        if "start" in node.nodes[0].extra or "end" in node.nodes[0].extra:
            wide_output = True
    value_lens = []
    values = []

    _type = None

    def __d(n):
        nonlocal value_lens, _type, values, debug
        value_lens.append(len(str(n.start)))
        value_lens.append(len(str(n.end)))
        if debug:
            values.append(str(n.start))
            values.append(str(n.end))
        if _type is None or _type == "int":
            if isinstance(n.start, int):
                _type = "int"
            else:
                _type = "float"
        for x in n.nodes:
            __d(x)

    __d(node)
    value_column_len = max(value_lens)
    return wide_output, value_column_len, _type


DEFAULT_FLOAT_DIGIT = 3
TREE_COL_DIS = 1


def _pp_tree(
    node,
    time_format=FORMAT_TIMESTAMP,
    file=None,
    _prefix="",
    _last=True,
    span=None,
    level=1,
    root_name=TREE_ROOT_NAME,
    wide_output=False,
    value_column_len=14,
    _type="float",
    _float_digit=DEFAULT_FLOAT_DIGIT,
    tree_extra=None,
):
    """pretty print in tree format"""
    if not tree_extra:
        tree_extra = {}
    _p = ""
    if wide_output:
        if node.extra:
            if time_format == FORMAT_TIMESTAMP:
                if _type == "float":
                    number_format = f"%{11+_float_digit}.{_float_digit}f"
                    _p += (
                        f"[{number_format%node.extra[EXTRA_START_TIME_KEY]} - {number_format%node.extra[EXTRA_END_TIME_KEY]}]"
                        + " " * 2
                    )
                else:
                    number_format = r"%14.0f"
                    _p += (
                        f"[{number_format%node.extra[EXTRA_START_TIME_KEY]} - {number_format%node.extra[EXTRA_END_TIME_KEY]}]"
                        + " " * 2
                    )
            elif time_format == FORMAT_DATETIME:
                _p += (
                    f"[{ts_pair_to_dt(node.extra[EXTRA_START_TIME_KEY], node.extra[EXTRA_END_TIME_KEY])}]"
                    + " " * 2
                )
            else:
                raise ValueError(f"wrong time format: {time_format}")
        else:
            if time_format == FORMAT_DATETIME:
                _p += " " * (42 + _float_digit * 2)
                # _p += f"[{number_format%0} - {number_format%0}]" + " " * 2
            else:
                _p += " " * (29 + _float_digit * 2)
    if isinstance(node.start, str):
        raise ValueError(f"wrong type of node.start: {node.start}")
    if len(str(node.start)) >= TS_LEN and float(node.start) > 1633299120 // 2:
        if time_format == FORMAT_TIMESTAMP:
            number_format = f"%{value_column_len-_float_digit}.{_float_digit}f"
            # _p += f"[{node.start:14.3f} ~ {node.end:14.3f}]" + " " * 2
            _p += (
                f"[{number_format%node.start} - {number_format%node.end}]"
                + " " * TREE_COL_DIS
            )
        elif time_format == FORMAT_DATETIME:
            _p += f"[{ts_pair_to_dt(node.start, node.end)}]" + " " * TREE_COL_DIS
        else:
            raise ValueError(f"wrong time format: {time_format}")
    else:
        # number_len = max(len(str(node.start)), len(str(node.end)))
        if _type == "float":
            # number_format = f"%{value_column_len+_float_digit}.{_float_digit}f"
            number_format = f"%{value_column_len}.{_float_digit}f"
            # print(number_format)
            _p += (
                f"[{number_format%node.start} - {number_format%node.end}]"
                + " " * TREE_COL_DIS
            )
        else:
            number_format = f"%{value_column_len}.0f"
            _p += (
                f"[{number_format%node.start} - {number_format%node.end}]"
                + " " * TREE_COL_DIS
            )
    if span is None:
        span = node.span() if level > 0 else 0
        branch = "🟢"
    else:
        branch = "l"
    if span != 0:
        _p += f" [{(node.span())*100/span:6.2f}%]" + " " * TREE_COL_DIS
    else:
        _p += " [000.00%]" + " " * TREE_COL_DIS
    _node_name = (
        f"root_{root_name}" if (is_none(node.name) and level == 1) else node.name
    )
    if node.extra:
        if wide_output:
            del node.extra[EXTRA_START_TIME_KEY], node.extra[EXTRA_END_TIME_KEY]
        contents = f"{_node_name}({node.span():6.4f})"  # TODO: no time to do it
        if node.extra:
            contents += f" ({str(node.extra)})"
    else:
        contents = f"{_node_name}({node.span():6.4f})"
    if level == 1 and tree_extra:
        # print("🤑" * 40, id(tree_extra))
        overhead_str, tree_extra_str = "", ""
        if "overhead" in tree_extra:
            overhead_str = f"[OH:{int(tree_extra['overhead'])}us]"
            del tree_extra["overhead"]
            del tree_extra["overhead_start"]
        if len(tree_extra) > 0:
            pretty_ = concise_dict(tree_extra)
            if pretty_:
                tree_extra_str = f"[{pretty_}]"
        prefix_length = len(_p + _prefix) + 2
        if tree_extra_str:
            contents += "\n" + " " * prefix_length + tree_extra_str
        if overhead_str:
            contents += "\n" + " " * prefix_length + overhead_str
    print(_p, _prefix, branch + "_" if _last else "|_", contents, sep="", file=file)
    _prefix += "   " if _last else "|  "
    child_count = len(node.nodes)
    for i, child in enumerate(node.nodes):
        _last = i == (child_count - 1)
        _pp_tree(
            child,
            time_format,
            file,
            _prefix,
            _last,
            span=span,
            level=level + 1,
            wide_output=wide_output,
            value_column_len=value_column_len,
            _type=_type,
        )


def concise_dict(d: dict):
    _d = {}
    for k, v in d.items():
        if isinstance(v, str) and len(v) > 20:
            _d[k] = "..." + v[-20:]
        elif not v:
            pass
        else:
            _d[k] = v
    return _d


def is_none(x) -> bool:
    return x is None or x.lower() == "none"


def dict_to_b64_str(d: dict):
    """d is a dictionary with strings/int/float inside"""
    b64 = base64.b64encode(bytes(str(d), "utf-8"))
    return b64.decode("utf-8")


@memoize
def b64_str_to_dict(s: str) -> dict:
    """d is a dictionary with strings/int/float inside"""
    try:
        d = base64.b64decode(s).decode("utf-8")
        r = ast.literal_eval(d)
        return r
    except Exception as e:
        return {}



import os
import hiq
from typing import *
import time
from hiq.memory import get_memory_mb
from hiq.hiq_utils import (
    func_args_handler,
    get_tau_id,
)
from hiq.utils import get_env_int

here = os.path.dirname(os.path.realpath(__file__))

HIQ_PADDLEOCR_CONF = hiq.hiq_utils.read_csv_to_list(f"{here}/paddleocr.conf")

class PaddleOcrHiQLatency(hiq.HiQSimple):
  
    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time],
        hiq_id_func: Callable = get_tau_id,
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
    ):
        extra_hiq_table += HIQ_PADDLEOCR_CONF
        hiq.HiQSimple.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=hiq_id_func,
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )
  
    def custom(s):
        s.o_paddle_run = hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run

        @s.inserter
        def paddle_run(self) -> bool:
            return s.o_paddle_run(self)

        hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run = paddle_run

    def custom_disable(s):
        hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run = s.o_paddle_run


class PaddleOcrHiQMemory(hiq.HiQMemory):
    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time, get_memory_mb],
        hiq_id_func: Callable = get_tau_id,
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
    ):
        extra_hiq_table += HIQ_PADDLEOCR_CONF
        hiq.HiQSimple.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=hiq_id_func,
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )

    def custom(s):
        s.o_paddle_run = hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run

        @s.inserter
        def paddle_run(self) -> bool:
            return s.o_paddle_run(self)

        hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run = paddle_run

    def custom_disable(s):
        hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run = s.o_paddle_run



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import contextvars
import gc
import os
import re
import time
from multiprocessing import resource_tracker, shared_memory
from typing import *

from cachetools import TTLCache, cached

from hiq.constants import *
from hiq.ddict import ddjson
from hiq.utils import get_env_bool

DEFAULT_HIQ_STATUS = get_env_bool("HIQ_ENABLED", True)
HIQ_STATUS_CACHED = get_env_bool("HIQ_STATUS_CACHED", False)


def __remove_shm_from_resource_tracker():
    """Monkey-patch multiprocessing.resource_tracker so SharedMemory won't be tracked

    More details at: https://bugs.python.org/issue38119
    """
    from multiprocessing import Process, resource_tracker
    from multiprocessing.shared_memory import SharedMemory

    def fix_register(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.register(self, name, rtype)

    resource_tracker.register = fix_register

    def fix_unregister(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.unregister(self, name, rtype)

    resource_tracker.unregister = fix_unregister

    if "shared_memory" in resource_tracker._CLEANUP_FUNCS:
        del resource_tracker._CLEANUP_FUNCS["shared_memory"]


def set_global_hiq_status(on=True, name="hiq", debug=False):
    """Set the global HiQ status. True is to enable HiQ, and False disable.

    >>> import hiq
    >>> hiq.get_global_hiq_status()
    True
    >>> hiq.set_global_hiq_status(0)
    😇 set global hiq to 0
    >>> hiq.get_global_hiq_status()
    False
    """
    try:
        c = shared_memory.ShareableList(name=name)
        c[0] = on
        if debug:
            print(f"😇 set global hiq to {on}")
    except FileNotFoundError:
        __remove_shm_from_resource_tracker()
        c = shared_memory.ShareableList([on], name=name)
        resource_tracker.unregister(name, "shared_memory")
        if debug:
            print(f"😤 set global hiq to {on}")
    except PermissionError as e:
        return False
    except Exception as e:
        raise e


def get_global_hiq_status(name="hiq", default=DEFAULT_HIQ_STATUS) -> bool:
    """Get the global HiQ status. True means HiQ is enabled, and False disabled.

    When `HIQ_STATUS_CACHED` is set as True, the global hiq status will be cached for 5 seconds. Otherwise, it will read shared memory.

    >>> import hiq
    >>> hiq.get_global_hiq_status()
    False
    >>> hiq.set_global_hiq_status(1)
    😇 set global hiq to 1
    >>> hiq.get_global_hiq_status()
    True
    """
    if HIQ_STATUS_CACHED:
        return _get_global_hiq_on_cached(name, default)
    return _get_global_hiq_on(name, default)


# cache data for no longer than 2 minutes
@cached(cache=TTLCache(maxsize=1024, ttl=GLOBAL_HIQ_TTL_S))
def _get_global_hiq_on_cached(name="hiq", default=DEFAULT_HIQ_STATUS):
    """
    time cost: 102.04315185546875 us.
    """
    try:
        return bool(shared_memory.ShareableList(name=name)[0])
    except FileNotFoundError as e:
        return default
    except Exception as e:
        raise e


def _get_global_hiq_on(name="hiq", default=DEFAULT_HIQ_STATUS):
    """
    time cost: 102.04315185546875 us.
    """
    try:
        return bool(shared_memory.ShareableList(name=name)[0])
    except (FileNotFoundError, PermissionError) as e:
        return default
    except Exception as e:
        raise e


class HiQStatusContext(object):
    """An HiQ context manager

    Inside HiQStatusContext, HiQ status is always enabled unless the status is changed in other processes(HiQ status is not guarded).

    >>> from hiq import HiQStatusContext
    >>> with HiQStatusContext():
            # HiQ will be enabled inside the `with` block, and reverted to original value out of the block
    >>> with HiQStatusContext(target_status_on=False):
            # HiQ will be disabled inside the `with` block, and reverted to original value out of the block
    """

    def __init__(self, target_status_on=True, debug=False):
        """Constructor of HiQStatusContext

        Args:
            target_status_on (bool, optional): set the target HiQ status you want to set in the context manager. Defaults to True.
            debug (bool, optional): print more information when debug is True. Defaults to False.
        """
        self.original_hiq_status = get_global_hiq_status()
        self.target_on = target_status_on
        self.set = False
        self.debug = debug

    def __enter__(self):
        if self.original_hiq_status != self.target_on:
            set_global_hiq_status(self.target_on, debug=self.debug)
            self.set = True

    def __exit__(self, type, value, traceback):
        if self.set:
            set_global_hiq_status(self.original_hiq_status, debug=self.debug)


def read_csv_to_list(file_path):
    res = []
    # CSV
    import csv

    try:
        with open(file_path, mode="r", encoding="utf-8") as csvfile:
            spamreader = csv.reader(
                csvfile,
                delimiter=",",
                quotechar='"',
                doublequote=True,
                skipinitialspace=True,
            )
            for row in spamreader:
                res.append(row)
        return res
    except Exception as e:
        return []


def get_hiq_table(hiq_table_or_path):
    if isinstance(hiq_table_or_path, str):
        if not os.path.exists(hiq_table_or_path):
            raise Exception(f"😓 file {hiq_table_or_path} doesn't exist.")
        res = []
        # JSON
        try:
            data = ddjson(hiq_table_or_path)
            for d in data:
                res.append([d.o_module, d.o_class, d.o_function, d.o_name])
            return res
        except ValueError as e:
            pass
        return read_csv_to_list(hiq_table_or_path)
    elif isinstance(hiq_table_or_path, list):
        return hiq_table_or_path
    else:
        raise Exception(f"bad data:{hiq_table_or_path}")


def _is_callable(a_: str):
    if not re.match(r"^[\w\.]+$", a_):
        raise Exception(f"wrong input {a_}")
    try:
        exec(f"_exist = callable({a_})", globals(), locals())
    except NameError:
        return 0
    return locals()["_exist"]


def silent_import(module_name: str):
    try:
        import importlib
        return importlib.import_module(module_name)
    except:
        return None


np = silent_import("numpy")
torch = silent_import("torch")
pandas = silent_import("pandas")


def __value_to_str(i, depth=1):
    if depth == 3:
        return "..."
    s = ""
    if i is None:
        s += "None"
    elif isinstance(i, bytes):
        s += f"bytes({len(i)})"
    elif np and isinstance(i, np.ndarray):
        s += f"ndarry{str(i.shape).replace(' ','')}"
    elif isinstance(i, dict):
        tmp = []
        for k, v in i.items():
            tmp.append(__value_to_str(v, depth+1))
        s += f"dict(k:{list(i.keys())},v:{tmp})"
    elif isinstance(i, list):
        tmp = []
        for idx, e in enumerate(i):
            tmp.append(__value_to_str(e, depth+1))
            if idx == 10:
                tmp.append('...')
                break
        s += f"list({len(i)},{','.join(tmp)})"
    elif torch and isinstance(i, torch.Tensor):
        s += f"tensor({str(i.shape).replace(' ','')})"
    elif pandas and isinstance(i, pandas.core.frame.DataFrame):
        s += f"pandas({str(i.shape).replace(' ','')})"
    else:
        if hasattr(i, "__len__"):
            s += f"{type(i).__name__}({str(len(i))})"
        else:
            s += f"{type(i).__name__}({i})"
    return s


def __func_arguments_str(args: List[Any]) -> str:
    """
    Tracing large data structure like arrays could be a performance killer. It will take a lot of CPU and some memory as well, and slow down the program.

    There are three ways to handle it:

    - Log a summary (default behavior): By default, if you log an array with size > 10,000, HiQ will only log the first 10,000 values, along with the shape.
    - Omit the array: You can also just choose not to log the array at all.
    - Manual transformation: .

    Args:
        args (List[Any]): the function arguments

    Returns:
        str : a string representation of the arguments
    """

    arg_list = []
    for i in args:
        arg_list.append(__value_to_str(i))
    return ",".join(arg_list)


def __func_kwargs_str(kwargs: Dict[str, Any]) -> str:
    r = {}
    for k, v in kwargs.items():
        r[k] = __value_to_str(v)
    return str(r)


def func_args_handler(x: Any, func_name=None) -> str:
    if isinstance(x, tuple):
        return __func_arguments_str(x)
    elif isinstance(x, dict):
        return __func_kwargs_str(x)
    raise ValueError(f"input has wrong type: {type(x)}")


def call_decorated(f: Callable, args=None, kwargs=None, tracing_type=TRACING_TYPE_HIQ):
    """Call with tracing type: HIQ or OCI

    HIQ is to build the HiQTree and then output or send to agent.
    OCI is to collect span and send data to OCI APM server.

    HIQ is the default way for monolithic application. OCI APM sends data in the same thread so it has longer latency but it is better for distributed applications.

    """
    if not args:
        args = {}
    if not kwargs:
        kwargs = {}
    if tracing_type == TRACING_TYPE_HIQ:
        return f(*args, **kwargs)
    elif tracing_type == TRACING_TYPE_OCI:
        from py_zipkin.zipkin import zipkin_span

        with zipkin_span(
                service_name=os.environ.get("SERVICE_NAME", "hiq"),
                span_name=f.__name__,
        ):
            return f(*args, **kwargs)
    elif tracing_type == TRACING_TYPE_OTM:
        from opentelemetry import trace

        tracer_name = os.environ.get("OTM_TRACER_NAME", "otm_hiq")
        tracer = trace.get_tracer(tracer_name)
        with tracer.start_as_current_span(f.__name__):
            return f(*args, **kwargs)
    elif tracing_type == TRACING_TYPE_PROM:
        from hiq.prometheus import get_summary

        with get_summary(f.__name__).time():
            return f(*args, **kwargs)

    else:
        return f(*args, **kwargs)


SERVER_NAME = "flask"
flask_dependency_ok = True


def get_tau_id(server_name=None):
    global flask_dependency_ok
    if server_name == SERVER_NAME and flask_dependency_ok:
        try:
            from flask import has_request_context, request

            tau_id = None
            if has_request_context():
                tau_id = request.environ.get("HTTP_X_REQUEST_ID")
            return tau_id
        except:
            flask_dependency_ok = False
            pass
    return None


# import cvpickle
class HiQIdGenerator(object):
    counter = 0

    def __init__(self):
        import itree

        self.context_var = contextvars.ContextVar("🆔")
        # cvpickle.register_contextvar(self.context_var, __name__)
        self.context_var.set(str(itree.time_us()) + str(HiQIdGenerator.counter % 10))
        HiQIdGenerator.counter += 1

    def __call__(self):
        return str(self.context_var.get())


class FlaskReqIdGenerator(object):
    def __call__(self):
        from flask import has_request_context, request

        return (
            request.environ.get("HTTP_X_REQUEST_ID")
            if has_request_context()
            else 'flask-req'
        )


get_global_hiq_status()

if __name__ == "__main__":
    start = time.perf_counter()
    tau_on = get_global_hiq_status()
    end = time.perf_counter()
    print(f"time cost: {(end - start) * 1e6}us, {tau_on}")
    for i in range(20):
        o = HiQIdGenerator()
        print(o())



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import logging
import random
import string

from flask import _app_ctx_stack, current_app, g, request
from py_zipkin import zipkin

from hiq.vendor_oci_apm import OciApmHttpTransport


class FlaskWithOciApm(object):
    """A Flask App's helper class with Oci-Apm supported by default

    To make it work, you need to set two environment variables: APM_BASE_URL, APM_PUB_KEY

    Example:

    .. highlight:: python
    .. code-block:: python

        import os
        import time

        from flask import Flask
        from flask_request_id_header.middleware import RequestID
        from hiq.server_flask_with_oci_apm import FlaskWithOciApm


        def create_app():
            app = Flask(__name__)
            app.config["REQUEST_ID_UNIQUE_VALUE_PREFIX"] = "hiq-"
            RequestID(app)
            return app


        app = create_app()
        amp = FlaskWithOciApm()
        amp.init_app(app)


        @app.route("/", methods=["GET"])
        def index():
            return "OK"


        if __name__ == "__main__":
            host = "0.0.0.0"
            port = int(os.getenv("PORT", "8080"))
            debug = False
            app.run(host=host, port=port, debug=debug)

    """

    def _gen_random_id(self):
        return "".join(random.choice(string.digits) for i in range(16))

    def __init__(
        self,
        app=None,
        sample_rate=100,
        timeout=1,
        transport_handler_=OciApmHttpTransport,
    ):
        self._exempt_views = set()
        self._sample_rate = sample_rate
        if app is not None:
            self.init_app(app)
        self._transport_handler = transport_handler_
        self._transport_exception_handler = None
        self._timeout = timeout

    def init_app(self, app):
        self.app = app
        app.before_request(self._before_request)
        app.after_request(self._after_request)
        self._disable = app.config.get(
            "ZIPKIN_DISABLE", app.config.get("TESTING", False)
        )
        return self

    def _should_use_token(self, view_func):
        return view_func not in self._exempt_views

    def _safe_headers(self, headers):
        self._headers = dict((k.lower(), v) for k, v in headers.__iter__())
        return self._headers

    def _before_request(self):
        if self._disable:
            return

        _app_ctx_stack.top._view_func = current_app.view_functions.get(request.endpoint)

        if not self._should_use_token(_app_ctx_stack.top._view_func):
            return

        safe_headers = self._safe_headers(request.headers)

        parent_span_id = safe_headers.get("x-b3-parentspanid")
        trace_id = safe_headers.get("x-b3-traceid") or self._gen_random_id()
        is_sampled = str(safe_headers.get("x-b3-sampled") or "0") == "1"

        flags = safe_headers.get("x-b3-flags")

        zipkin_attrs = zipkin.ZipkinAttrs(
            trace_id=trace_id,
            span_id=self._gen_random_id(),
            parent_span_id=parent_span_id,
            flags=flags,
            is_sampled=is_sampled,
        )

        span = zipkin.zipkin_span(
            service_name=self.app.name,
            span_name="{0}.{1}".format(request.endpoint, request.method),
            transport_handler=self._transport_handler,
            sample_rate=self._sample_rate,
            zipkin_attrs=zipkin_attrs,
        )
        g._zipkin_span = span
        g._zipkin_span.start()

    def exempt(self, view):
        view_location = "{0}.{1}".format(view.__module__, view.__name__)
        self._exempt_views.add(view_location)
        return view

    def _after_request(self, response):
        if self._disable:
            return response
        if not hasattr(g, "_zipkin_span"):
            return response
        try:
            g._zipkin_span.stop()
        except Exception as e:
            logging.warning("Unable to stop zipkin span:{}".format(e))

        return response

    def create_http_headers_for_new_span(self):
        if self._disable:
            return dict()
        return zipkin.create_http_headers_for_new_span()

    def logging(self, **kwargs):
        logging.warning(
            "This method has been deprecated, " "please call `update_tags` instead."
        )
        self.update_tags(**kwargs)

    def update_tags(self, **kwargs):
        if all(
            [hasattr(g, "_zipkin_span"), g._zipkin_span, g._zipkin_span.logging_context]
        ):
            g._zipkin_span.logging_context.binary_annotations_dict.update(kwargs)


def child_span(f):
    def decorated(*args, **kwargs):
        import flask

        span = zipkin.zipkin_span(
            service_name=flask.current_app.name,
            span_name=f.__name__,
        )
        kwargs["span"] = span
        with span:
            val = f(*args, **kwargs)
            span.update_binary_annotations(
                {
                    "function_args": args,
                    "function_returns": val,
                }
            )
            return val

    return decorated



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import json
import os
import re

from collections.abc import Mapping
from collections import UserDict
from collections import namedtuple


CLS_NAME = "GAMMA_DYNAMIC_DICT"
PREFIX = "o_"

__all__ = ["ddlist", "ddjson", "ddyaml"]


def ddlist(keys=[], values=[], name=CLS_NAME):
    mapping = dict(zip(keys, values))
    return _mapper(mapping, _nt_name=name)


def ddjson(data_or_path: str, name=CLS_NAME):
    if os.path.exists(data_or_path):
        with open(data_or_path, "r", encoding="UTF-8") as f:
            return _mapper(json.load(f), _nt_name=name)
    return _mapper(json.loads(data_or_path), _nt_name=name)


def ddyaml(data_or_path: str, name=CLS_NAME):
    if not os.path.exists(data_or_path):
        return _mapper(data_or_path, _nt_name=name)
    import yaml

    with open(data_or_path, "r", encoding="UTF-8") as file:
        data = yaml.load(file, yaml.FullLoader)
        return _mapper(data, _nt_name=name)


def _mapper(mapping, _nt_name=CLS_NAME):
    """Convert mappings to namedtuple recursively."""
    if isinstance(mapping, Mapping) and not isinstance(mapping, AsDict):
        for key, value in list(mapping.items()):
            mapping[key] = _mapper(value)
        return namedtuple_wrapper(_nt_name, **mapping)
    elif isinstance(mapping, list):
        return [_mapper(item) for item in mapping]
    return mapping


def namedtuple_wrapper(_nt_name, **kwargs):
    import keyword

    n = {}
    for k, v in kwargs.items():
        nk = re.sub(r"[^0-9a-zA-Z_]+", "_", k)
        n[PREFIX + nk] = v
    wrap = namedtuple(_nt_name, n)
    return wrap(**n)


class AsDict(UserDict):
    """placeholder"""


def ignore(mapping):
    if isinstance(mapping, Mapping):
        return AsDict(mapping)
    elif isinstance(mapping, list):
        return [ignore(item) for item in mapping]
    return mapping


def is_namedtuple_instance(x):
    _type = type(x)
    bases = _type.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(_type, "_fields", None)
    if not isinstance(fields, tuple):
        return False
    return all(type(i) == str for i in fields)


def _reducer(obj):
    if isinstance(obj, dict):
        return {key: _reducer(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [_reducer(value) for value in obj]
    elif is_namedtuple_instance(obj):
        return {key: _reducer(value) for key, value in obj._asdict().items()}
    elif isinstance(obj, tuple):
        return tuple(_reducer(value) for value in obj)
    else:
        return obj



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import time
import requests
import os

from py_zipkin import Encoding
from py_zipkin.zipkin import zipkin_span


def get_oci_apm_endpoint(APM_BASE_URL="", key=""):
    endpoint = os.environ.get("APM_BASE_URL", APM_BASE_URL)
    key = os.environ.get("APM_PUB_KEY", key)  # "JL6DVW2YBYYPA6G53UG3ZNNEN642AKPA"
    API_VER = os.environ.get("APM_API_VER", "20200101")

    final_endpoint = f"{endpoint}/{API_VER}/observations/public-span?dataFormat=zipkin&dataFormatVersion=2&dataKey={key}"
    if not key:
        raise ValueError("🦉 Please setup environment variable `APM_PUB_KEY`")
    if not endpoint:
        raise ValueError("🦉 Please setup environment variable `APM_BASE_URL`")
    return final_endpoint


class OciApmHttpTransport(object):
    """a zipkin transport_handler class to emit traces to Oracle APM

    To make it work, you need to set two environment variables: `APM_BASE_URL`, `APM_PUB_KEY`

    Example:

    .. highlight:: python
    .. code-block:: python

        import time

        from py_zipkin import Encoding
        from py_zipkin.zipkin import zipkin_span
        from hiq.vendor_oci_apm import OciApmHttpTransport


        def fun():
            with zipkin_span(
                service_name="hiq_test_apm",
                span_name="fun_test",
                transport_handler=OciApmHttpTransport,
                encoding=Encoding.V2_JSON,
                binary_annotations={"mode": "sync"},
                sample_rate=100,
            ):
                time.sleep(5)
                print("hello")


        if __name__ == "__main__":
            fun()

    """

    def __init__(s, encoded_span, apm_url=None, debug=False):
        if apm_url is None:
            apm_url = get_oci_apm_endpoint()
        s.encoded_span = encoded_span
        s.apm_url = apm_url
        s.debug = debug
        s.call()

    def call(s):
        if s.debug:
            start = time.monotonic()
        res = requests.post(
            s.apm_url,
            data=s.encoded_span,
            headers={"Content-Type": "application/json"},
        )
        if s.debug:
            print(f"😎 {len(s.encoded_span)=}")
            end = time.monotonic()
            post_latency_cost = (end - start) * 1e6
            print(f"⏰ {post_latency_cost=}us, {res.status_code=}")


class HiQOciApmContext(zipkin_span):
    """Logs a root zipkin span with HiQ support

    Example:

    .. highlight:: python
    .. code-block:: python

        import time

        from hiq.vendor_oci_apm import HiQOciApmContext


        def fun():
            with HiQOciApmContext(
                service_name="hiq_test_apm", span_name="fun_test",
            ):
                time.sleep(5)
                print("hello")


        if __name__ == "__main__":
            fun()
    """

    def __init__(
        self,
        service_name="hiq_service_name",
        span_name="hiq_span_name",
        binary_annotations={"mode": "sync"},
        sample_rate=100,
    ):
        """Constructor of HiQOciApmContext class

        Args:
            service_name (str, optional): The name of the called service. Defaults to "hiq_service_name".
            span_name (str, optional): root span name. Defaults to "hiq_span_name".
            binary_annotations (dict, optional): dict of str -> str span attrs. Defaults to {"mode": "sync"}.
            sample_rate (int, optional): Rate at which to sample; 0.0 - 100.0. If passed-in zipkin_attrs have is_sampled=False and the sample_rate param is > 0, a new span will be generated at this rate. This means that if you propagate sampling decisions to downstream services, but still have sample_rate > 0 in those services, the actual rate of generated spans for those services will be > sampling_rate. Defaults to 100.
        """
        zipkin_span.__init__(
            self,
            service_name=service_name,
            span_name=span_name,
            transport_handler=OciApmHttpTransport,
            encoding=Encoding.V2_JSON,
            binary_annotations=binary_annotations,
            sample_rate=sample_rate,
        )

    def __enter__(self):
        zipkin_span.__enter__(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        super().__exit__(exc_type, exc_value, exc_traceback)



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
#

import ast
import builtins
import inspect
import io
import json
import os
import shutil
import subprocess
import sys
import traceback
from datetime import datetime
from functools import wraps
from time import monotonic
from typing import Callable, List, Tuple, Union


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def __gamma_split(s: str, keep_delim=False) -> List[str]:
    results = []
    word = ""
    x = 0
    for c in s:
        if x == 0:
            if c in ("\t", "\n"):
                continue
        if c == "'":
            if x == 0:
                x = 1
                if keep_delim:
                    word += "'"
                continue
            elif x == 1:
                if keep_delim:
                    word += "'"
                results.append(word)
                word = ""
                x = 0
                continue
        elif c == " ":
            if x == 1:
                pass
            elif x == 0:
                if word:
                    results.append(word)
                    word = ""
                continue
        word += c
    if word:
        results.append(word)
    return results


def execute_cmd(
    command: str,
    split=True,
    verbose=True,
    check=False,
    shell=False,
    timeout=600,
    stderr_log=None,
    debug=False,
    keep_delim=False,
    env=None,
    error_file=None,
    append_error=False,
) -> Union[str, List[str]]:
    """
    If verbose is true, print out input.
    If check is true, and the process exits with a non-zero exit code, a CalledProcessError exception will be raised.
    Attributes of that exception hold the arguments, the exit code, and stdout and stderr if they were captured.
    stderr_log could be: stderr_log=open("/tmp/gamma.error.log", "a+")
    https://docs.python.org/3/library/subprocess.html#subprocess.run
    """
    error_file = error_file or f"/tmp/error.{random_str()}.log"
    if stderr_log is None:
        stderr_log = open(error_file, "w", encoding="utf8")
    try:
        commands = __gamma_split(command, keep_delim)
        if verbose:
            print(f"🏃‍♂️ command: {' '.join(commands)}, error_file: {error_file}")
            print(f"🏃‍♂️ commands: {commands}")
        if env is None:
            result = subprocess.run(
                commands,
                stdout=subprocess.PIPE,
                stderr=stderr_log,
                check=check,
                shell=shell,
                timeout=timeout,
            )
        else:
            result = subprocess.run(
                commands,
                stdout=subprocess.PIPE,
                stderr=stderr_log,
                check=check,
                shell=shell,
                timeout=timeout,
                env=env,
            )
        if os.path.exists(error_file) and os.path.getsize(error_file) and debug:
            print("☠️ error:", read_file(error_file, binary_mode=True, by_line=False))
        if split:
            ret = result.stdout.decode("utf-8").strip().split("\n")
        else:
            ret = result.stdout.decode("utf-8").strip()
        if append_error and os.path.exists(error_file) and os.path.getsize(error_file) > 0:
            if isinstance(ret, str):
                ret += read_file(error_file, by_line=False)
            else:
                ret.append(read_file(error_file, by_line=False))
        return ret
    except subprocess.TimeoutExpired:
        print(f"🏃‍♂️ command timed-out after {timeout} seconds: {command}")
    finally:
        if stderr_log and hasattr(stderr_log, "close"):  # don't want to import io :-)
            stderr_log.close()
        if os.path.exists(error_file) and not debug:
            if os.path.getsize(error_file) == 0:
                os.unlink(error_file)
                if verbose:
                    print(f"{error_file} is empty and removed")
            elif verbose:
                print("*" * 80)
                print(read_file(error_file, by_line=False))
                print("*" * 80)


def utc_to_pst(utc: str, format_="%Y-%m-%dT%H:%M:%S+0000") -> str:
    """utc2pst("2021-06-19T20:32:18+0000")"""
    from dateutil import tz

    from_zone = tz.gettz("UTC")
    to_zone = tz.gettz("America/Los_Angeles")
    utc = datetime.strptime(utc, format_)
    utc = utc.replace(tzinfo=from_zone)
    central = utc.astimezone(to_zone)
    return str(central)


def ts_to_dt(timestamp: float) -> str:
    if timestamp < 1633279464 // 2:
        # print(f"warning: {timestamp} is not a timestamp")
        return str(timestamp)
    try:
        dt_object = datetime.fromtimestamp(timestamp)
        return dt_object.strftime(r"%Y-%m-%d %H:%M:%S.%f")
    except OverflowError as e:
        print(f"error: timestamp {timestamp} cased overflow")
        return ""


def ts_pair_to_dt(t1: float, t2: float) -> str:
    d1 = datetime.fromtimestamp(t1)
    s1 = d1.strftime(r"%Y-%m-%d %H:%M:%S.%f")
    p1 = s1.split(" ")
    d2 = datetime.fromtimestamp(t2)
    s2 = d2.strftime(r"%Y-%m-%d %H:%M:%S.%f")
    p2 = s2.split(" ")
    if p1[0] == p2[0]:
        return f"{p1[0]} {p1[1]} - {p2[1]}"
    return f"{s1} - {s2}"


def _get_args_spec(args_spec) -> Tuple[str, str]:
    x = args_spec.args
    varkw = args_spec.varkw
    if args_spec.defaults is None:
        if varkw:
            x.append(f"**{varkw}")
        return ",".join(x), ",".join(x)
    y = list(args_spec.defaults)
    y = ["^"] * (len(x) - len(y)) + y
    z = list(zip(x, y))
    r = list(
        map(
            lambda e: e[0]
            if e[1] == "^"
            else (f"{e[0]}='{e[1]}'" if isinstance(e[1], str) else f"{e[0]}={e[1]}"),
            z,
        )
    )
    if varkw:
        r.append(f"**{varkw}")
        x.append(f"**{varkw}")
    return ",".join(r), ",".join(x)


# File System related


def read_file(
    file_path: str,
    binary_mode: bool = False,
    by_line: bool = True,
    filter_func: Callable = None,
    as_json=False,
    bytes_as_string=True,
    raise_=False,
    strip=False,
) -> Union[List[str], bytes]:
    """A handy function to read file

    Args:
        file_path (str): file path
        raise_ (bool): populate exception when it happens, for instance, when it's False, in case the file doesn't exist, return None
        strip (bool): strip the string output or not?

    Don't use this for super large file.

    Examples:

    .. highlight:: python
    .. code-block:: python

        >>> data = read_file(f"{here}/o.json")
        >>> pp(data)
        ['{',
        '    "eventKey": "repo:refs_changed",',
        '    "date": "2021-09-21T04:52:51+0000",',
        '    "actor": {',
        '        "name": "fuhengwu",',
        '        "emailAddress": "fuheng.wu@oracle.com",',
        '        "id": 937189,',
        '        "displayName": "fuheng wu",',
        '        "active": true,',
        '        "slug": "fuhengwu",',
        '        "type": "NORMAL",',
        '        "links": {',
        '            "self": [{ "href": '
        '"https://www.github.com/users/fuhengwu" }]',
        '        }',
        '    }',
        '}']
        >>> data = read_file(f"{here}/o.json", binary_mode=True)
        >>> pp(data)
        (b'{\\n    "eventKey": "repo:refs_changed",\\n    "date": "2021-09-21T04:52:51+'
        b'0000",\\n    "actor": {\\n        "name": "fuhengwu",\\n        "emailAddress":'
        b' "fuheng.wu@oracle.com",\\n        "id": 937189,\\n        "displayName": "F'
        b'uheng Wu",\\n        "active": true,\\n        "slug": "fuhengwu",\\n        "t'
        b'ype": "NORMAL",\\n        "links": {\\n            "self": [{ "href": "https'
        b'://www.github.com/users/fuhengwu" }]\\n        }\\n    }\\n}')
        >>> data = read_file(f"{here}/o.json", as_json=True)
        >>> pp(data)
        {'actor': {'active': True,
                'displayName': 'fuheng wu',
                'emailAddress': 'fuheng.wu@oracle.com',
                'id': 937189,
                'links': {'self': [{'href': 'https://www.github.com/users/fuhengwu'}]},
                'name': 'fuhengwu',
                'slug': 'fuhengwu',
                'type': 'NORMAL'},
        'date': '2021-09-21T04:52:51+0000',
        'eventKey': 'repo:refs_changed'}
    """
    if not os.path.isfile(file_path):
        if not raise_:
            return None
        raise ValueError(f"input path {file_path} is not a file")
    # assert by_line != as_json, "Can not read file by line with as_json"
    if as_json:  # give as_json higher priority
        by_line = False
    res = b"" if binary_mode else []
    with open(file_path, "rb" if binary_mode else "r") as file:
        if binary_mode:
            res = file.read()
        else:
            res = file.read()
            if by_line:
                res = res.splitlines()
                if filter_func:
                    res = list(filter(filter_func, res))
    if as_json:
        res = json.loads(res)
    if not binary_mode and isinstance(res, (bytes,)) and bytes_as_string:
        res = res.decode("utf-8")
    if isinstance(res, str) and strip:
        return res.strip()
    return res


def write_file(file_path, data, as_owner=None, as_group=None, append=False, mod_="644"):
    """write data into file. if file_path is just a file name, the file will be created in current directory"""
    if os.path.exists(file_path) and as_owner:
        execute_cmd(f"sudo chown {as_owner} {file_path}", timeout=10, debug=True)
    if "/" in file_path:
        ensure_folder(file_path)
    if append:
        execute_cmd(f"touch {file_path}")
    with open(file_path, "ab" if append else "wb") as file:
        file.write(data.encode("utf-8"))
        if as_owner:
            execute_cmd(f"sudo chown {as_owner} {file_path}", timeout=10, debug=True)
        if as_group:
            execute_cmd(f"sudo chgrp {as_group} {file_path}", timeout=10, debug=True)
        if mod_ != "644":
            execute_cmd(f"chmod {mod_} {file_path}", timeout=10, debug=True)


def ensure_folder(path_str: str):
    """ensure there is a folder for the path_str, the input is supposed to be a FILE path"""
    try:
        if not os.path.isdir(path_str):
            path_str = os.path.dirname(path_str)
        if not os.path.exists(path_str):
            os.makedirs(path_str)
        return path_str
    except PermissionError as e:
        print(f"permission error: {path_str}")
        raise e


def download_from_http(
    uri, local_file_path, display=False, enable_proxy=True, auth=None
) -> str:
    """auth=(user, password)"""
    import requests
    import urllib3

    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    try:
        proxies = {}
        if enable_proxy:
            proxies = get_proxies()
        if not (auth is None or (isinstance(auth, (tuple, list)) and len(auth) == 2)):
            raise ValueError(f"auth information is wrong: {auth}")
        response = requests.get(
            uri, verify=False, stream=True, proxies=proxies, auth=auth
        )
        if response.status_code >= 400:
            raise IOError(
                response.status_code,
                f"Error retrieving {uri}. {response.status_code}: {response.text}",
            )
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        if not display:
            response.raw.decode_content = True
            with open(local_file_path, "wb") as f:
                shutil.copyfileobj(response.raw, f)
        else:
            from tqdm import tqdm

            if total_size_in_bytes > 0:
                block_size = 1024 * 500  # 500 KiB
                with tqdm(
                    total=total_size_in_bytes, unit="iB", unit_scale=True
                ) as progress_bar:
                    file_obj = io.BytesIO()
                    for data in response.iter_content(block_size):
                        progress_bar.update(len(data))
                        file_obj.write(data)
                if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
                    raise ValueError(
                        f"💀 error happens when downloading {uri} to {local_file_path}"
                    )
            if total_size_in_bytes <= 0:
                response = requests.get(uri, verify=False)
                file_obj = io.BytesIO()
                file_obj.write(response.content)
            file_obj.seek(0, 0)
            with open(local_file_path, "wb") as out:
                out.write(file_obj.read())
        return local_file_path
    except Exception as e:
        print(f"🦉 error: {str(e)}")
        traceback.print_exc(file=sys.stdout)
        raise e


def random_str(length_of_string=12):
    import random
    import string

    letters_and_digits = string.ascii_lowercase + string.digits
    random_string = ""
    for _ in range(length_of_string):
        random_string += random.choice(letters_and_digits)
    return random_string


class SilencePrint(object):
    def __enter__(self):
        self.old_print = builtins.print
        builtins.print = lambda *args, **kwargs: None

    def __exit__(self, exc_type, exc_value, exc_tb):
        builtins.print = self.old_print


def get_proxies() -> dict:
    proxies = {
        "http": os.environ.get("http_proxy", ""),
        "https": os.environ.get("https_proxy", ""),
        "ftp": os.environ.get("ftp_proxy", ""),
        "no_proxy": os.environ.get("no_proxy", ""),
    }
    return proxies


def _check_overhead(f, *args, **kwargs):
    @wraps(f)
    def __y(s, *args, **kwargs):
        start = monotonic()
        r = f(s, *args, **kwargs)
        s.check_oh_counter += 1
        s.overhead_us += int((monotonic() - start) * 1e6)
        return r

    return __y


def get_env_bool(x, default=None) -> bool:
    """Get bool type environment variable

    These values will be treated as True:

        non-empty string, non-zero numeric


    >>> import os
    >>> from hiq.utils import get_env_bool
    >>> get_env_bool('hello')
    False
    >>> os.environ["hello"]=""
    >>> get_env_bool('hello')
    False
    >>> os.environ["hello"]="1"
    >>> get_env_bool('hello')
    True
    >>> os.environ["hello"]="0"
    >>> get_env_bool('hello')
    False
    >>> os.environ["hello"]="true"
    >>> get_env_bool('hello')
    True

    """
    t = os.environ.get(x, default)
    if isinstance(t, str) and t:
        try:
            return bool(ast.literal_eval(t))
        except:
            return True
    return bool(t)


def get_env_int(x, default=0) -> int:
    return int(os.environ.get(x, default))


def get_env_float(x, default=0) -> float:
    return float(os.environ.get(x, default))


def is_hiqed(fun: Callable, fun_name: str) -> bool:
    """check if this function has been registered in HiQ system

    The function could be one of:

        full_qualified_name="<method 'read' of '_io.TextIOWrapper' objects>", fun_name='read'

        full_qualified_name='<function main at 0x7f7f2eaf6040>', fun_name='main'

        full_qualified_name='<built-in function read>', fun_name='read'

    Args:
        fun (Callable): original function
        fun_name (str): function name

    Returns:
        bool: is this function has been registered in HiQ
    """
    full_qualified_name = str(fun)
    # print(f"🐵 {full_qualified_name=}, {fun_name=}")
    return (
        f"{fun_name} at" not in full_qualified_name
        and f"<method '{fun_name}' of" not in full_qualified_name
        and not full_qualified_name.startswith("<built-in function")
    )


def lmk_data_handler(data: dict = {}, pid=os.getpid()) -> str:
    """convert raw data of Log Monkey King to the format we want to log

    Args:
        data (dict, optional): a data dictionary. If not empty, the keys should be: `id_`, `name`, `value`, `extra`, `is_start`. Defaults to {}.
        pid ([type], optional): LMK process id. Defaults to os.getpid().

    Returns:
        str: a log entry as a data string
    """
    id_ = data["id_"]
    name = data["name"][2:]
    value = data["value"]
    extra = data["extra"]
    is_start = data["is_start"]

    monkey = "🐵" if is_start else "🙈"
    if "time" in extra:
        data = f"{ts_to_dt(extra['time'])} - [{id_}] [🆔 {pid}] {monkey} [{name}] {value} {extra}"
    else:
        if extra:
            data = f"{ts_to_dt(value)} - [{id_}] [🆔 {pid}] {monkey} [{name}] {extra}"
        else:
            data = f"{ts_to_dt(value)} - [{id_}] [🆔 {pid}] {monkey} [{name}]"
    return data


def get_home():
    from pathlib import Path

    return str(Path.home())


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SingletonBase(metaclass=SingletonMeta):
    pass


def memoize(func):
    cache = dict()

    def __x(*args):
        if args in cache:
            return cache[args]
        result = func(*args)
        cache[args] = result
        return result

    return __x


def memoize_first(func):
    cache = dict()

    def __x(*args):
        if args[0] in cache:
            return cache[args[0]]
        result = func(*args)
        cache[args[0]] = result
        return result

    return __x


def get_cpu_info(debug=False):
    cpuset = os.sched_getaffinity(0)
    if debug:
        print(f"{len(os.sched_getaffinity(0))} CPUs are available: {cpuset}")
    return cpuset


def pretty_time_delta(seconds):
    sign_string = "-" if seconds < 0 else ""
    seconds = abs(int(seconds))
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days > 0:
        return "%s%dd%dh%dm%ds" % (sign_string, days, hours, minutes, seconds)
    elif hours > 0:
        return "%s%dh%dm%ds" % (sign_string, hours, minutes, seconds)
    elif minutes > 0:
        return "%s%dm%ds" % (sign_string, minutes, seconds)
    else:
        return "%s%ds" % (sign_string, seconds)


def create_gantt_chart_time(data: List[str], fig_path=None):
    """Plot Gantt-chart for HiQ Latency Tree"""
    from hiq import tree
    import pandas as pd
    import matplotlib.pyplot as plt
    import numpy as np
    import random
    from matplotlib.patches import Patch

    if isinstance(data, str):
        data = [data]

    # Return DataFrame of Data
    def create_df(data):
        # Traverse tree to get name and time-stamps
        def dfs(result, pass_name, num, node):
            if pass_name == "":
                if node.start == float("inf"):
                    # start time
                    s = ""
                    # end time
                    e = ""
                else:
                    # start time
                    s = datetime.fromtimestamp(float(node.start))
                    # end time
                    e = datetime.fromtimestamp(float(node.end))
                result.append(
                    [
                        str(pass_name) + str(node.name.strip("_")) + "." + str(num),
                        str(node.name.strip("_")),
                        s,
                        e,
                    ]
                )
            else:
                if node.start == float("inf"):
                    # start time
                    s = ""
                    # end time
                    e = ""
                else:
                    # start time
                    s = datetime.fromtimestamp(float(node.start))
                    # end time
                    e = datetime.fromtimestamp(float(node.end))
                result.append(
                    [
                        str(pass_name)
                        + "."
                        + str(node.name.strip("_"))
                        + "."
                        + str(num),
                        str(node.name.strip("_")),
                        s,
                        e,
                    ]
                )
            if node.nodes:
                for nd in node.nodes:
                    num += 1
                    if str(node.name) == "None":
                        pn = str(pass_name)
                    else:
                        if pass_name == "":
                            pn = str(pass_name) + str(node.name.strip("_"))
                        else:
                            pn = str(pass_name) + "." + str(node.name.strip("_"))
                    result = dfs(result, pn, num, nd)
                return result
            else:
                return result

        all_result = []
        total_data = len(data)
        for i in range(total_data):
            # Create tree
            t = tree.Tree(data[i])
            # Extract data from Tree
            pass_name = "Call-" + str(i + 1)
            dfs_result = []
            dfs_result = dfs(dfs_result, pass_name, 0, t.root)
            dfs_result = dfs_result[1:]

            req = []
            req.append(pass_name)
            req.append("Requests")
            req_start = min([s[2] for s in dfs_result])
            req_end = max([e[3] for e in dfs_result])
            req.append(req_start)
            req.append(req_end)
            dfs_result.append(req)

            all_result.append(dfs_result)

        # Generate random Hex color code
        def get_color():
            hex_number = [
                "#" + "".join([random.choice("ABCDEF0123456789") for i in range(6)])
            ]
            return hex_number[0]

        # Create dfs
        def create_dfs_from_list(req_result):
            df = pd.DataFrame(req_result, columns=["Name", "Group", "Start", "End"])

            proj_start = df.Start.min()

            # Difference is in seconds
            df["start_num"] = (df.Start - proj_start).dt.total_seconds()

            df["end_num"] = (df.End - proj_start).dt.total_seconds()

            df["time_start_to_end"] = df.end_num - df.start_num

            c_dict = {}
            for i in df["Group"].unique():
                c_dict[i] = get_color()

            # Attach color to the group
            def color(row, cc_dict=c_dict):
                return cc_dict[row["Group"]]

            df["color"] = df.apply(color, axis=1)

            return [df, c_dict, proj_start]

        all_dfs = []
        for r in all_result:
            all_dfs.append(create_dfs_from_list(r))

        return all_dfs

    all_dfs = create_df(data)

    #### Create Plots ####
    fl = len(all_dfs)
    fig, ax_plots = plt.subplots(1, fl, figsize=(fl * 10, 7))

    for i in range(fl):
        if fl == 1:
            ax = ax_plots
        else:
            ax = ax_plots[i]
        df = all_dfs[i][0]
        c_dict = all_dfs[i][1]
        proj_start = all_dfs[i][2]

        ax.barh(df.Name, df.time_start_to_end, left=df.start_num, color=df.color)

        legend_elements = [
            Patch(facecolor=c_dict[i], label=i.strip("_")) for i in c_dict
        ]
        ax.legend(handles=legend_elements)

        number_of_vertical_lines = 14
        interval = (df.end_num.max() + 1) / number_of_vertical_lines
        xticks = np.arange(0, df.end_num.max() + 1, interval)
        xticks_labels = (
            pd.date_range(proj_start, end=df.End.max(), periods=len(xticks))
            .floor("ms")
            .format(".%3f")[1:]
        )
        xticks_minor = np.arange(0, df.end_num.max() + 1, 1)

        ax.tick_params(rotation=80, axis="x")
        ax.set_xticks(xticks)
        ax.set_xticks(xticks_minor, minor=True)
        ax.set_xticklabels(xticks_labels[::1])
        ax.title.set_text("Call-" + str(i + 1))
        ax.set_xlabel("Timestamp")
        ax.set_ylabel("Function")
        ax.grid()

    fig.tight_layout()
    if fig_path is None:
        fig_path = f"/tmp/tim_{random_str()}.png"
    ensure_folder(fig_path)
    plt.savefig(fig_path)
    if os.path.exists(fig_path):
        return fig_path


def create_gantt_chart_memory(data: List[str], fig_path=None):
    """Plot Gantt-chart for HiQ Memory Tree"""
    from hiq import tree
    import pandas as pd
    import matplotlib.pyplot as plt
    import numpy as np
    import random
    from matplotlib.patches import Patch

    if isinstance(data, str):
        data = [data]

    # Return DataFrame of Data
    def create_df(data):
        # Traverse tree to get name and time-stamps
        def dfs(result, pass_name, num, node, r_min, r_max):
            if pass_name == "":
                if node.name == "None":
                    s, e, m = "", "", ""
                else:
                    if node.extra["start"] == float("inf"):
                        s, e, m = "", "", ""
                    else:
                        s = datetime.fromtimestamp(float(node.extra["start"]))
                        e = datetime.fromtimestamp(float(node.extra["end"]))
                        m = float(node.end) - float(node.start)
                        r_min = min(r_min, float(node.start))
                        r_max = max(r_max, float(node.end))
                result.append(
                    [
                        str(pass_name) + str(node.name.strip("_")) + "." + str(num),
                        str(node.name.strip("_")),
                        s,
                        e,
                        m,
                    ]
                )
            else:
                if node.name == "None":
                    s, e, m = "", "", ""
                else:
                    if node.extra["start"] == float("inf"):
                        s, e, m = "", "", ""
                    else:
                        s = datetime.fromtimestamp(float(node.extra["start"]))
                        e = datetime.fromtimestamp(float(node.extra["end"]))
                        m = float(node.end) - float(node.start)
                        r_min = min(r_min, float(node.start))
                        r_max = max(r_max, float(node.end))
                result.append(
                    [
                        str(pass_name)
                        + "."
                        + str(node.name.strip("_"))
                        + "."
                        + str(num),
                        str(node.name.strip("_")),
                        s,
                        e,
                        m,
                    ]
                )
            if node.nodes:
                for nd in node.nodes:
                    num += 1
                    if str(node.name) == "None":
                        pn = str(pass_name)
                    else:
                        if pass_name == "":
                            pn = str(pass_name) + str(node.name.strip("_"))
                        else:
                            pn = str(pass_name) + "." + str(node.name.strip("_"))
                    result, r_min, r_max = dfs(result, pn, num, nd, r_min, r_max)
                return result, r_min, r_max
            else:
                return result, r_min, r_max

        all_result = []
        total_data = len(data)
        for i in range(total_data):
            # Create tree
            t = tree.Tree(data[i])
            # Extract data from Tree
            pass_name = "Call-" + str(i + 1)
            dfs_result = []
            dfs_result, r_min, r_max = dfs(
                dfs_result, pass_name, 0, t.root, float("inf"), 0
            )
            dfs_result = dfs_result[1:]

            req = []
            req.append(pass_name)
            req.append("Requests")
            req_start = min([s[2] for s in dfs_result])
            req_end = max([e[3] for e in dfs_result])
            req_memory = r_max - r_min
            req.append(req_start)
            req.append(req_end)
            req.append(req_memory)
            dfs_result.append(req)

            all_result.append(dfs_result)

        # Generate random Hex color code
        def get_color():
            hex_number = [
                "#" + "".join([random.choice("ABCDEF0123456789") for i in range(6)])
            ]
            return hex_number[0]

        # Create dfs
        def create_dfs_from_list(req_result):
            df = pd.DataFrame(
                req_result, columns=["Name", "Group", "Start", "End", "Memory"]
            )
            proj_start = df.Start.min()
            # Difference is in seconds
            df["start_num"] = (df.Start - proj_start).dt.total_seconds()
            df["end_num"] = (df.End - proj_start).dt.total_seconds()
            df["time_start_to_end"] = df.end_num - df.start_num
            c_dict = {}
            for i in df["Group"].unique():
                c_dict[i] = get_color()
            # Attach color to the group
            def color(row, cc_dict=c_dict):
                return cc_dict[row["Group"]]

            df["color"] = df.apply(color, axis=1)
            return [df, c_dict, proj_start]

        all_dfs = []
        for r in all_result:
            all_dfs.append(create_dfs_from_list(r))
        return all_dfs

    all_dfs = create_df(data)

    #### Create Plots ####
    fl = len(all_dfs)
    fig, ax_plots = plt.subplots(1, fl, figsize=(fl * 10, 7))

    for i in range(fl):
        if fl == 1:
            ax = ax_plots
        else:
            ax = ax_plots[i]
        df = all_dfs[i][0]
        c_dict = all_dfs[i][1]
        proj_start = all_dfs[i][2]

        ax.barh(df.Name, df.time_start_to_end, left=df.start_num, color=df.color)
        # Display memory
        for idx, row in df.iterrows():
            ax.text(
                row.end_num + 0.1,
                idx,
                format(float(row.Memory), ".3f"),
                va="center",
                alpha=0.8,
            )
        legend_elements = [
            Patch(facecolor=c_dict[i], label=i.strip("_")) for i in c_dict
        ]
        ax.legend(handles=legend_elements)

        number_of_vertical_lines = 14
        interval = (df.end_num.max() + 1) / number_of_vertical_lines
        xticks = np.arange(0, df.end_num.max() + 1, interval)
        xticks_labels = (
            pd.date_range(proj_start, end=df.End.max(), periods=len(xticks))
            .floor("ms")
            .format(".%3f")[1:]
        )
        xticks_minor = np.arange(0, df.end_num.max() + 1, 1)

        ax.tick_params(rotation=80, axis="x")
        ax.set_xticks(xticks)
        ax.set_xticks(xticks_minor, minor=True)
        ax.set_xticklabels(xticks_labels[::1])
        ax.title.set_text("Call-" + str(i + 1))
        ax.set_xlabel("Timestamp")
        ax.set_ylabel("Function")
        ax.grid()

    fig.tight_layout()
    if fig_path is None:
        fig_path = f"/tmp/mem_{random_str()}.png"
    ensure_folder(fig_path)
    plt.savefig(fig_path)
    if os.path.exists(fig_path):
        return fig_path


if __name__ == "__main__":
    import requests

    print(pretty_time_delta(730.2312341))
    args = _get_args_spec(inspect.getfullargspec(requests.get))
    print(args)
    data = read_file(f"{get_home()}/.bashrc", by_line=False, binary_mode=False)
    print(data)

    mtree = """t1^get_memory_mb,,0,108,5,0,1e-07,0#%n1*[None,inf,inf,0$0#[__predict,2146.82421875,2146.37890625,0$55#{'start': 1643569736.1171584, 'end': 1643569747.438287}[__txt,2146.82421875,2146.12890625,0$56#{'start': 1643569736.1535447, 'end': 1643569744.0208213}[__det,2146.82421875,2146.546875,0$55#{'start': 1643569736.2086039, 'end': 1643569738.931958}[__ort,2146.82421875,2146.546875,0$56#{'start': 1643569736.3555777, 'end': 1643569738.8348844}]][__rec,2146.546875,2146.12890625,0$55#{'start': 1643569739.5116146, 'end': 1643569743.115841}[__ort,2146.546875,2146.12890625,0$56#{'start': 1643569740.1161213, 'end': 1643569742.6137362}]]][_ort_sess,2146.37890625,2146.41796875,0$54#{'start': 1643569745.352344, 'end': 1643569745.630854}]]]"""
    print(create_gantt_chart_memory([mtree]))

    tau = """t1^time,,0,108,5,1,1e-07,47#{'overhead_start': 14461965, 'overhead': 39488}%n1*[None,inf,inf,0$0#[__predict,1643569736.1171172,1643569747.4377651,0$29#{'start': 1643569736.1171584}[__pdf,1643569736.1183336,1643569736.1530828,0$0#][__txt,1643569736.1535318,1643569744.0208154,0$0#[__sdc,1643569736.1536727,1643569736.16426,0$0#[__ort,1643569736.160469,1643569736.1639962,0$0#]][__res,1643569736.1644254,1643569736.2080786,0$0#[__ort,1643569736.1652586,1643569736.2076447,0$0#]][__det,1643569736.20859,1643569738.931932,0$0#[__ort,1643569736.3555562,1643569738.8348587,0$0#]][__cls,1643569738.9405437,1643569739.24126,0$0#[__ort,1643569738.9429545,1643569739.0299811,0$0#][__ort,1643569739.0350616,1643569739.127901,0$0#][__ort,1643569739.132973,1643569739.2317488,0$0#][__ort,1643569739.2328234,1643569739.241071,0$0#]][__sty,1643569739.241476,1643569739.5114026,0$0#[__ort,1643569739.243572,1643569739.4119606,0$0#][__ort,1643569739.4144793,1643569739.4317007,0$0#][__ort,1643569739.4342513,1643569739.5082533,0$0#][__ort,1643569739.5091777,1643569739.5111573,0$0#]][__rec,1643569739.511604,1643569743.115833,0$0#[__ort,1643569739.5133343,1643569739.7487059,0$0#][__ort,1643569739.7528548,1643569740.0487385,0$0#][__ort,1643569740.1160934,1643569742.6137102,0$0#][__ort,1643569742.6292832,1643569743.1126142,0$0#]][__rec,1643569743.11609,1643569743.116233,0$0#][__det,1643569743.1164277,1643569743.1422083,0$0#[__ort,1643569743.1169982,1643569743.1402473,0$0#]][__det,1643569743.1426477,1643569743.1468668,0$0#[__ort,1643569743.143132,1643569743.145506,0$0#]][__det,1643569743.1472528,1643569743.2162328,0$0#[__ort,1643569743.1480312,1643569743.2121801,0$0#]][__det,1643569743.216631,1643569743.2421668,0$0#[__ort,1643569743.2174027,1643569743.2351496,0$0#]][__det,1643569743.242663,1643569743.2523344,0$0#[__ort,1643569743.2434154,1643569743.2501614,0$0#]][__det,1643569743.3165214,1643569743.329657,0$0#[__ort,1643569743.3176396,1643569743.3263106,0$0#]][__det,1643569743.3300874,1643569743.337189,0$0#[__ort,1643569743.330739,1643569743.3349724,0$0#]][__det,1643569743.337554,1643569743.342569,0$0#[__ort,1643569743.3380115,1643569743.3409102,0$0#]][__det,1643569743.3429003,1643569743.3468688,0$0#[__ort,1643569743.3433163,1643569743.345672,0$0#]][__det,1643569743.3472311,1643569743.3520138,0$0#[__ort,1643569743.3476806,1643569743.3503542,0$0#]][__det,1643569743.3524318,1643569743.3581665,0$0#[__ort,1643569743.3529246,1643569743.3563924,0$0#]][__det,1643569743.3585193,1643569743.4109883,0$0#[__ort,1643569743.3589113,1643569743.409384,0$0#]][__det,1643569743.411437,1643569743.4156177,0$0#[__ort,1643569743.4118354,1643569743.4143524,0$0#]][__det,1643569743.4159608,1643569743.420702,0$0#[__ort,1643569743.4164329,1643569743.418999,0$0#]][__det,1643569743.4210744,1643569743.4252756,0$0#[__ort,1643569743.4214988,1643569743.4240358,0$0#]][__det,1643569743.4256399,1643569743.4298625,0$0#[__ort,1643569743.426082,1643569743.4286163,0$0#]][__det,1643569743.4302363,1643569743.4343379,0$0#[__ort,1643569743.4306252,1643569743.433071,0$0#]][__det,1643569743.4346728,1643569743.4385982,0$0#[__ort,1643569743.435092,1643569743.4374073,0$0#]][__det,1643569743.4389238,1643569743.4435484,0$0#[__ort,1643569743.4393814,1643569743.4419074,0$0#]][__det,1643569743.4438791,1643569743.4482865,0$0#[__ort,1643569743.4443324,1643569743.4470842,0$0#]][__det,1643569743.4486613,1643569743.454831,0$0#[__ort,1643569743.4492812,1643569743.452746,0$0#]][__det,1643569743.4554029,1643569743.5328221,0$0#[__ort,1643569743.4574807,1643569743.5185628,0$0#]][__det,1643569743.5338244,1643569743.5492887,0$0#[__ort,1643569743.5352147,1643569743.5429738,0$0#]][__det,1643569743.5498576,1643569743.6286142,0$0#[__ort,1643569743.5520155,1643569743.618229,0$0#]][__det,1643569743.6297498,1643569743.7098956,0$0#[__ort,1643569743.632189,1643569743.6457608,0$0#]][__det,1643569743.710675,1643569743.733005,0$0#[__ort,1643569743.7126865,1643569743.7247958,0$0#]][__det,1643569743.7337465,1643569743.8100944,0$0#[__ort,1643569743.7356975,1643569743.7473083,0$0#]][__det,1643569743.8111055,1643569743.8347492,0$0#[__ort,1643569743.813039,1643569743.8244736,0$0#]][__det,1643569743.835479,1643569743.84312,0$0#[__ort,1643569743.836138,1643569743.8404937,0$0#]][__det,1643569743.843564,1643569743.8490357,0$0#[__ort,1643569743.844158,1643569743.8473823,0$0#]][__det,1643569743.8493845,1643569743.913976,0$0#[__ort,1643569743.8504245,1643569743.9103267,0$0#]][__det,1643569743.9145095,1643569743.9196005,0$0#[__ort,1643569743.9151065,1643569743.9183722,0$0#]][__det,1643569743.919938,1643569743.924156,0$0#[__ort,1643569743.9203613,1643569743.9229279,0$0#]][__det,1643569743.9244788,1643569743.9358187,0$0#[__ort,1643569743.9257007,1643569743.932356,0$0#]][__det,1643569743.9362285,1643569743.9403598,0$0#[__ort,1643569743.9366255,1643569743.939164,0$0#]][__det,1643569743.94068,1643569743.9454722,0$0#[__ort,1643569743.9411945,1643569743.9442341,0$0#]][__det,1643569743.9458015,1643569743.9520783,0$0#[__ort,1643569743.9464548,1643569743.9499493,0$0#]][__det,1643569743.9524298,1643569744.0121005,0$0#[__ort,1643569743.953075,1643569744.0093017,0$0#]][__det,1643569744.0125837,1643569744.0203052,0$0#[__ort,1643569744.0133524,1643569744.0177798,0$0#]]][__kvlayout,1643569744.15425,1643569745.3451543,0$0#[__ort,1643569744.1545331,1643569745.3447654,0$0#]][_ort_sess,1643569745.3523283,1643569745.6308331,0$0#][__tbl,1643569745.6312604,1643569745.9097774,0$0#[__ort,1643569745.6314573,1643569745.8510344,0$0#]]]]"""
    print(create_gantt_chart_time([tau]))



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import os
import json
import syslog
import time
import requests
import urllib3
from hiq.utils import read_file
import queue


urllib3.disable_warnings(urllib3.exceptions.SecurityWarning)
# Suppress warnings about deprecated use of certificate common name.
# RFC-2818 deprecated the CN field in server certificates, but it's still
# the most common way to indicate the server hostname. The openssl tool
# makes it ridiculously difficult to use the SAN field (the recommended
# alternative), and support for SAN across various systems seems sketchy,
# at best. The Python maintainers have decided that it's worth a nasty
# runtime warning. We don't agree with that, so I've added this line.
import datetime


def __emit_log(sev, msg):
    """
    Emit a log to stdout. sev should be a syslog error level; msg can be any
    text string.
    """
    print(datetime.datetime.now().isoformat(), sev, msg)


class HttpMetricsClient(object):
    """
    HttpMetricsClient is a generic class for transmitting metrics to any metrics server by HTTP

    This includes the config needed to create and submit metrics, along with the requests.Session we use to send the data.

    Examples:

    .. highlight:: python
    .. code-block:: python

        from hiq.http_metric_client import HttpMetricsClient as Client
        from time import monotonic

        client = Client(url=...)
        METRIC_NAME = "hiq.predict"
        metric = Client.metric_calc_delta(f"{METRIC_NAME}.success", start_time=monotonic())
        metrics_queue.put_nowait(metric)
        #...
        retry_count=1
        client.gauge_metric("operation_retry_count", retry_count)
        client.submit_metrics_queue()

    """

    @staticmethod
    def metric_calc_delta(metric, start_time):
        """Send a metric with a time measurement"""
        end_time = time.monotonic()
        delta = round((end_time - start_time) * 1000)
        metric_data = {
            "name": metric,
            "datapoints": [
                {"timestamp": int(round(time.time() * 1000)), "value": delta}
            ],
        }
        return metric_data

    def __init__(
        self,
        url,
        ad_longform=None,
        metrics_queue=None,
        trusted_cert=None,
        project="hiq",
        timeout=5,
    ):
        """Constructor

        Parameters:
            url: The metrics data server URL
            ad_longform: the long-form name of the availability domain of the style like `eu-frankfurt-ad-1`
            metrics_queue: a Python queue.Queue object used for queuing metrics
            trusted_cert: The filename of the root certificate for authenticating the server, if needed
        """
        self.conf = {}
        self.conf["hostname"] = read_file("/etc/hostname", by_line=False, strip=True)
        region = read_file("/etc/region", by_line=False, strip=True)
        self.conf["region"] = region if region else "none"
        self.conf["ad_longform"] = ad_longform
        self.conf["url"] = url
        self.metrics_queue = metrics_queue if metrics_queue else queue.Queue()
        if not trusted_cert or not os.path.exists(trusted_cert):
            trusted_cert = "/etc/pki/ca-trust/extracted/openssl/ca-bundle.trust.crt"
        self.trusted_cert = trusted_cert if os.path.exists(trusted_cert) else None
        self.timeout = timeout
        self.project = project

    def timing_metric_data(self, metric_data):
        """Send a list of metrics server

        :param metric_data list of metrics
        """
        info = self.wrap_metric_data(metric_data=metric_data)

        try:
            resp = requests.put(
                f'{self.conf["url"]}aggregation',
                timeout=self.timeout,
                json=info,
                verify=self.trusted_cert,
            )
        except requests.exceptions.ConnectionError as err:
            __emit_log(syslog.LOG_ERR, f"network connection error: {err}")
        else:
            if resp.status_code < 200 or resp.status_code >= 300:
                __emit_log(
                    syslog.LOG_ERR, f"server status error: {resp} {info} {resp.text}"
                )

    def wrap_metric_data(self, metric_data) -> dict:
        """
        package up metrics in the data structure metrics server wants

        :param metric_data: info['metrics'] => list of metrics
        :return: info (metric dict/json)
        """
        info = {
            "project": self.project,
            "hostname": self.conf["hostname"],
            "region": self.conf["region"],
            "availabilityDomain": self.conf["ad_longform"],
            "metrics": metric_data,
        }
        if fleet := os.getenv("T2_FLEET", ""):
            info["fleet"] = fleet
        return info

    def submit_metrics_queue(self):
        """drain metrics_queue and submit the metrics to server"""
        metric_data = list(self.metrics_queue.queue)
        if not metric_data:
            msg = "metric_data must not be empty!"
            raise ValueError(msg)
        self.timing_metric_data(metric_data=metric_data)
        # handy for peeking at request to server and metric_data
        # __emit_log(syslog.LOG_DEBUG,
        #          f'json: {json.dumps(self.wrap_metric_data(metric_data))}')

    def gauge_metric(self, metric, value):
        """emit a gauge metric with the given name and value."""
        metric_data = {
            "name": metric,
            "datapoints": [
                {"timestamp": int(round(time.time() * 1000)), "value": value}
            ],
        }
        self.metrics_queue.put_nowait(metric_data)

    def timing_metric(self, metric: str, start_time: int):
        """
        convenience method to calculate a time delta for a metric and also put it in the metrics_queue

        :param metric: metric name
        :param start_time: start time in millisecond unit
        """
        metric_data = self.metric_calc_delta(metric, start_time)
        self.metrics_queue.put_nowait(metric_data)

    def get(self, query, params=None):
        """
        method for mock metrics server API during testing, not applicable to prod
        """
        resp = requests.get(
            f'{self.conf["url"]}{query}',
            params=params,
            timeout=self.timeout,
            verify=self.trusted_cert,
        )
        try:
            return json.loads(resp.text), resp.headers
        except json.JSONDecodeError:
            msg = (
                "Failed to parse query response: "
                + f'{self.conf["url"]}{query} '
                + f"{repr(params)} "
                + f"(response code = {str(resp.status_code)})"
            )
            raise IOError(msg)


if __name__ == "__main__":
    o = HttpMetricsClient(url="", ad_longform="")



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

__author__ = None

import collections
import gc
import inspect
import io
import os
import re
import sys
import time
import traceback
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import *

import hiq
import itree
from hiq.constants import *
from hiq.hiq_utils import (
    _is_callable,
    call_decorated,
    func_args_handler,
    get_global_hiq_status,
    get_hiq_table,
    get_tau_id,
    HiQIdGenerator,
)
from hiq.jack import Jack
from hiq.monkeyking import LogMonkeyKing
from hiq.tree import Tree
from hiq.utils import _check_overhead, _get_args_spec, is_hiqed
from hiq.memory import get_memory_mb

here = os.path.dirname(os.path.realpath(__file__))


class HiQException(Exception):
    pass


class HiQBase(itree.ForestStats, LogMonkeyKing):
    __metaclass__ = ABC
    __no_none_key__ = False
    __slots__ = ()

    def __init__(
        sf,
        hiq_table_or_path: Union[str, List[Iterable[str]]] = [],
        metric_funcs: List[Callable] = [time.time],
        hiq_id_func: Callable = get_tau_id,
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table: List[str] = [],
        attach_timestamp=False,
        extra_metrics: Set[ExtraMetrics] = set(),  # EXTRA_METRIC_FILE
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
        *args,
        **kwargs,
    ):
        """constructor of ABC HiQBase

        Args:
            hiq_table_or_path (Union[str, List[Iterable[str]]], optional): this is an HiQ Conf, please refer to *HiQ Core Concepts* section in the HiQ documentation.Defaults to [].
            metric_funcs (List[Callable], optional): simple metric function with empty argument. Defaults to [time.time].
            hiq_id_func (Callable, optional): a callable to generate unique id for tau, the hiq map. Defaults to hiq.hiq_utils.get_tau_id.
            func_args_handler(Callable, optional): a callable to convert function args/kwargs into a string. Defaults to hiq.hiq_utils.func_args_handler.
            target_path (str, optional): the directory of the target code. Defaults to None.
            max_hiq_size (int, optional): the max size of hiq map. if the number is exceeded, tree will be sent to LMK. Defaults to 30.
            verbose (bool, optional): when verbose is true, more information will be recorded, like the full stack trace of exception will be recorded in HiQ tree node. Defaults to False.
            fast_fail (bool, optional): when it is true, raise exception to the upper level, don't swallow exceptions. Defaults to True.
            tpl (str, optional): hiq tpl path. Defaults to None.
            extra_hiq_table (List[str], optional): a list of string to decide to include metrics other than latency. Defaults to [].
            attach_timestamp (bool, optional): for non-time/non-latency metric tree, should we attach start, end timestamps in the extra dictionary? Defaults to False.
            extra_metrics (set, optional): metrics to track in `extra` field of HiQ node. The value could be EXTRA_METRIC_ARGS, EXTRA_METRIC_FILE or EXTRA_METRIC_FUNC and it is normally used in development environment. Defaults to set().
        """
        super().__init__()
        LogMonkeyKing.__init__(sf, lmk_path, lmk_handler, lmk_logger)
        if verbose:
            print("process 🆔 {}".format(os.getpid()))
        if target_path and target_path not in sys.path:
            sys.path.append(target_path)
        sf.tau = collections.defaultdict(dict)
        sf.count = 0
        sf.max_hiq_size = max_hiq_size
        hiq_quadruple = get_hiq_table(hiq_table_or_path)
        # for t in extra_hiq_table:
        #    hiq_quadruple.append(t)
        hiq_quadruple += extra_hiq_table
        sf._verify_input(hiq_quadruple)
        sf.get_tau_id = hiq_id_func
        sf.get_func_args = func_args_handler
        sf.verbose = verbose
        sf.fast_fail = fast_fail
        sf.attach_timestamp = attach_timestamp
        sf.extra_metrics = extra_metrics
        sf.__load_hiq_tpl(tpl)
        sf.metric_funcs = metric_funcs
        sf.__load_extra_metrics(extra_hiq_table)
        sf.enable_hiq()
        sf.check_oh_counter = 0

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.disable_hiq()

    def __load_hiq_tpl(sf, tpl):
        if tpl and os.path.exists(tpl):
            sf.load_tpl(tpl)
            return
        tpl = f"{here}/{TAU_TPL_LOC}"
        if os.path.exists(tpl):
            sf.load_tpl(tpl)
            return
        tpl = f"{here}/{TAU_PKL_LOC}"
        if os.path.exists(tpl):
            sf.itree_tpl = hiq.mod("pickle").load(open(tpl, "rb"))
        if not sf.itree_tpl:
            raise Exception(f"🈚️ empty tau tpl. tpl:{tpl}")

    def __load_extra_metrics(sf, extra_hiq_table):
        if TAU_TABLE_DIO_RD in extra_hiq_table:
            sf.metric_funcs.append(sf.get_dio_bytes_r)
        if TAU_TABLE_DIO_WT in extra_hiq_table:
            sf.metric_funcs.append(sf.get_dio_bytes_w)
        if HIQ_TABLE_SIO_RD in extra_hiq_table:
            sf.metric_funcs.append(sf.get_sio_bytes_r)
        if HIQ_TABLE_SIO_WT in extra_hiq_table:
            sf.metric_funcs.append(sf.get_sio_bytes_w)
        if TAU_TABLE_NIO_GET in extra_hiq_table:
            sf.metric_funcs.append(sf.get_nio_bytes_r)
        if TAU_TABLE_NIO_POS in extra_hiq_table:
            sf.metric_funcs.append(sf.get_nio_bytes_w)

    def _verify_input(sf, hiq_quadruple):
        sf.hiq_quadruple = []
        tag_names = set()
        for i in hiq_quadruple:
            i = list(i)
            if len(i) > 0 and i[0].startswith("#"):
                continue
            if len(i) != 4 or not isinstance(i[3], str) or len(i[3]) < 1:
                raise ValueError("🦉 tau table must have exact 4 columns")
            i[3] = (
                (HIQ_FUNC_PREFIX + i[3])
                if len(i[3]) >= 2 and i[3][:2] != HIQ_FUNC_PREFIX
                else i[3]
            )
            if i[3] in tag_names:
                raise ValueError(f"🦉 duplicated tag: {i[3]} in tau input!")
            if not re.match(r"^\w+$", i[3]):
                raise ValueError(
                    "🦉 only alpha numerics are allowed in tag but got: {i[3]}!"
                )
            tag_names.add(i[3])
            sf.hiq_quadruple.append(i)

    def set_extra_metrics(sf, extra_metrics: Iterable[ExtraMetrics]):
        """set extra metric information so that these information will enter the span node"""
        sf.extra_metrics = extra_metrics

    def get_tree(s, func: Callable, extra=None) -> Tree:
        """find the hiq tree, and attach extra on the tree if necessary
        Note: the extra information here will enter into tree level. To pass extra into node level, use `Tree.start()`.
        """
        if not extra:
            extra = {}
        if s.get_tau_id is None:
            s.get_tau_id = HiQIdGenerator()
        req_id = s.get_tau_id()
        if not req_id and HiQBase.__no_none_key__:
            print(f"warning: tau id is {req_id}")
        fn = func.__name__
        if req_id not in s.tau or fn not in s.tau[req_id]:
            if fn in [KEY_LATENCY]:
                s.tau[req_id][fn] = Tree(
                    extra=extra,
                    tid=fn,
                    queue_lmk=s.queue_lmk,
                )
            else:
                s.tau[req_id][fn] = Tree(
                    extra=extra,
                    tid=fn,
                    monotonic=False,
                    queue_lmk=s.queue_lmk,
                )
        if extra:
            s.tau[req_id][fn].extra.update(extra)
        r = s.tau[req_id][fn]
        if len(s.tau) > s.max_hiq_size + 1:
            d = s.tau.pop(list(s.tau.keys())[0])
            if hasattr(s, "send_trees_to_jack"):
                s.send_trees_to_jack(d)
        return r

    @_check_overhead
    def pre_processing(s, f_name, args=(), kwargs=None, tree_extra=None):
        """collect extra information for node"""
        if not kwargs:
            kwargs = {}
        if not tree_extra:
            tree_extra = {}
        log_monkey_file = os.environ.get("LOG_MONKEY_FILE", False)
        node_extra = {}
        if ExtraMetrics.ARGS in s.extra_metrics:
            if args:
                node_extra["args"] = s.get_func_args(args, f_name)
            if kwargs:
                node_extra["kwargs"] = s.get_func_args(kwargs, f_name)
        if ExtraMetrics.FILE in s.extra_metrics or log_monkey_file:
            caller = inspect.stack()[3]
            node_extra["file"] = f"{caller.filename}:{caller.lineno}"
        if ExtraMetrics.FUNC in s.extra_metrics:
            caller = inspect.stack()[3]
            node_extra["function"] = caller.function
        for func in s.metric_funcs:
            tmp = s.get_tree(func, extra=tree_extra)
            if func.__name__ == KEY_LATENCY:
                if "overhead_start" not in tmp.extra:
                    tmp.extra["overhead_start"] = s.overhead_us
                if EXTRA_START_TIME_KEY in node_extra:
                    del node_extra[EXTRA_START_TIME_KEY]
            else:
                if s.attach_timestamp:
                    node_extra[EXTRA_START_TIME_KEY] = time.time()
            an_extra = deepcopy(node_extra) if node_extra else {}
            # if s.overhead_us == 0:
            #    print(f"🤠,{func=},{id(tmp)=},{id(an_extra)=}")
            tmp.start(f_name, func(), extra=an_extra)

    @_check_overhead
    def post_processing(s, f_name):
        extra = {}
        for func in s.metric_funcs:
            if func.__name__ == KEY_LATENCY:
                if EXTRA_END_TIME_KEY in extra:
                    del extra[EXTRA_END_TIME_KEY]
            else:
                if s.attach_timestamp:
                    extra[EXTRA_END_TIME_KEY] = time.time()
            s.get_tree(func).end(f_name, func(), extra=deepcopy(extra) if extra else {})
        #########################################################

    @_check_overhead
    def update_overhead(s):
        # check tracing overhead
        for func in s.metric_funcs:
            if func.__name__ == KEY_LATENCY:
                tmp = s.get_tree(func)
                if tmp.just_before_ending_root():
                    tmp.extra["overhead"] = s.overhead_us - tmp.extra["overhead_start"]
                    # print(
                    #    f"👿, {func=},{id(tmp)=},{id(tmp.extra)=},"
                    #    f"{tmp.extra['overhead_start']=},{tmp.extra['overhead']=}"
                    # )

    @_check_overhead
    def handle_exception(s, f_name, e):
        """handle exception from calling the original function

        `fast_fail` means raising exception to the upper level, don't swallow exceptions

        Args:
            f_name (string): original function name
            e (Exception): the exception raised from the original function

        Raises:
            e: the original exception
        """
        node_extra = {KEY_EXC_SUM: e}
        if s.verbose:
            sio = io.StringIO()
            traceback.print_exc(file=sio)
            trace_msg = sio.getvalue()
            node_extra[KEY_EXC_TRA] = trace_msg
        if s.fast_fail:
            for func in s.metric_funcs:
                if func.__name__ == KEY_LATENCY:
                    if EXTRA_END_TIME_KEY in node_extra:
                        del node_extra[EXTRA_END_TIME_KEY]
                else:
                    if s.attach_timestamp:
                        node_extra[EXTRA_END_TIME_KEY] = time.time()
                tmp = s.get_tree(func)  # TODO: we can add exception here
                if func.__name__ == KEY_LATENCY and tmp.just_before_ending_root():
                    tmp.extra["overhead"] = (
                        s.overhead_us - tmp.extra["overhead_start"]
                    ) / len(s.metric_funcs)
                tmp.end(f_name, func(), extra=deepcopy(node_extra))
            raise e

    def inserter(s, f, *_args, **_kwargs):
        """If function signature has no positional argument, we can use this inserter directly."""

        def __x(*args, **kwargs):
            if not get_global_hiq_status():
                # print(f"👻 tau is disabled: {f.__name__}")
                # s.disable_hiq()
                return f(*args, **kwargs)
            # print(f"{f_name=}")
            f_name = f.__name__
            s.pre_processing(f_name, args, kwargs)
            result = None
            try:
                result = call_decorated(
                    f,
                    args,
                    kwargs,
                    tracing_type=os.environ.get("TRACE_TYPE", TRACING_TYPE_HIQ),
                )
            except Exception as e:
                s.handle_exception(f_name, e)
            s.post_processing(f_name)
            s.update_overhead()
            return result

        return __x

    def inserter_with_extra(s, extra=None, *_args, **_kwargs):
        def wrap(f):
            f_name = f.__name__
            tracing_type = os.environ.get("TRACE_TYPE", TRACING_TYPE_HIQ)

            def __x(*args, **kwargs):
                if not get_global_hiq_status():
                    print("👻 hiq is disabled (extra): {f_name}")
                    # s.disable_hiq()
                    return f(*args, **kwargs)
                s.pre_processing(f_name, args, kwargs, tree_extra=extra)
                result = None
                try:
                    result = call_decorated(f, args, kwargs, tracing_type=tracing_type)
                except Exception as e:
                    s.handle_exception(f_name, e)
                s.post_processing(f_name)
                s.update_overhead()
                return result

            return __x

        return wrap

    def enable_hiq(s, reset_trace=False):
        """Enable HiQ tracing

        Args:
            s (HiQBase): self object of HiQBase
            reset_trace (bool, optional): will the `s.tau` data structure will be reset to empty? Defaults to False.

        Returns:
            HiQBase: the current object itself
        """
        if s.enabled:
            return
        try:
            if reset_trace:
                s.reset()
            if not get_global_hiq_status():
                print("🌚 global hiq switch is off")
                return s
            s.custom()
            for m, c, f, t in s.hiq_quadruple:
                s._h(m, c, f, t)
            s.enabled = True
        except ModuleNotFoundError as e:
            print(f"🦉 this module should be in the sys path. error: {e}")
        return s

    def disable_hiq(s, reset_trace=False):
        """Disable HiQ tracing

        Args:
            s (HiQBase): self object of HiQBase
            reset_trace (bool, optional): will the `s.tau` data structure will be reset to empty? Defaults to False.
        """
        if not s.enabled:
            return
        for module_name, class_name, func_name, tag in s.hiq_quadruple:
            class_func = f"{class_name}.{func_name}"
            if not class_name:
                class_func = func_name
            try:
                d = f"hiq.mod('{module_name}').{class_func} = s.o_{tag}"
                itree.exe(d, locals())
            except TypeError as e:
                s.__disable_c(module_name, class_name, func_name, tag)
        s.custom_disable()
        s.enabled = False
        if reset_trace:
            s.reset()

    def enable_c(s, module_name: str, class_name: str, func_name: str, new_func):
        _m = hiq.mod(module_name)
        _class = getattr(_m, class_name)
        _func_name = func_name
        _referents = gc.get_referents(_class.__dict__)[0]
        _original_val = _referents.get(_func_name, None)
        _referents[_func_name] = new_func
        _c = hiq.mod(EXTERNAL_C_MODULE)
        _c.pythonapi.PyType_Modified(_c.py_object(_class))
        return _original_val

    def __disable_c(s, module_name: str, class_name: str, func_name: str, tag: str):
        _m = hiq.mod(module_name)
        _class = getattr(_m, class_name)
        _func_name = func_name
        _referents = gc.get_referents(_class.__dict__)[0]
        d = f"_referents['{_func_name}'] = s.o_{tag}"
        itree.exe(d, locals())
        _c = hiq.mod(EXTERNAL_C_MODULE)
        _c.pythonapi.PyType_Modified(_c.py_object(_class))

    def reset(s):
        """reset the hiq map data structure to empty

        Args:
            s (HiQBase): self object of HiQBase
        """
        s.tau = collections.defaultdict(dict)
        s.check_oh_counter = 0

    def _h(s, module_name, class_name, func_name, tag: str = ""):
        try:
            class_func = f"{class_name}.{func_name}"
            if not class_name:
                class_func = func_name
            d = f"m_ = hiq.mod('{module_name}').{class_func}"
            itree.exe(d, locals())
            assert hasattr(locals()["m_"], "__call__"), "a callable is needed"
            if not tag:
                tag += str(s.count)
            # already defined in the custom
            if _is_callable(tag):
                d = f"""hiq.mod('{module_name}').{class_func} = {tag}"""
                itree.exe(d, locals())
            else:
                itree.exe(f"s.o_{tag} = m_", locals())
                spec_with_default, spec_wo_default = _get_args_spec(
                    inspect.getfullargspec(locals()["m_"])
                )
                assert (
                    re.match(r"^[\w\.]+$", tag)
                    and ";" not in spec_with_default
                    and ";" not in spec_wo_default
                    and ";" not in module_name
                    and ";" not in class_func
                ), "no hacker"
                signature = spec_with_default
                if spec_with_default:
                    signature += ","

                if signature.startswith("**"):
                    signature = "s=s," + signature
                elif ",**" in signature:
                    i = signature.find(",**")
                    signature = signature[:i] + ",s=s" + signature[i:]
                else:
                    signature += "s=s"
                if signature.endswith(","):
                    signature = signature[:-1]
                if "." in class_func:
                    _class_name = class_func.split(".")[0]
                    _funct_name = class_func.split(".")[1]
                else:
                    _class_name = "🦉"
                    _funct_name = class_func
                if not s.itree_tpl:
                    raise Exception("empty tau tpl")
                d = s.itree_tpl.format(
                    tag=tag,
                    signature=signature,
                    spec_wo_default=spec_wo_default,
                    module_name=module_name,
                    _class_name=_class_name,
                    _funct_name=_funct_name,
                )
                d = d.replace(".🦉", "")
                if tag == HIQ_FUNC_PREFIX + HIQ_FUNC_DIO_RD:
                    d = d.replace("🐚", "s.dio_bytes_r += len(_r)")
                elif tag == HIQ_FUNC_PREFIX + HIQ_FUNC_DIO_WT:
                    d = d.replace("🐚", "s.dio_bytes_w += len(_r)")
                elif tag == HIQ_FUNC_PREFIX + HIQ_FUNC_SIO_RD:
                    d = d.replace("🐚", "s.sio_bytes_r += len(_r)")
                elif tag == HIQ_FUNC_PREFIX + HIQ_FUNC_SIO_WT:
                    d = d.replace("🐚", "s.sio_bytes_w += len(_r)")
                elif tag == HIQ_FUNC_PREFIX + HIQ_FUNC_NIO_GET:
                    d = d.replace(
                        "🐚",
                        "s.nio_bytes_r += len(_r.content) if (_r and _r.status_code < 400) else 0",
                    )
                elif tag == HIQ_FUNC_PREFIX + HIQ_FUNC_NIO_POS:
                    d = d.replace(
                        "🐚",
                        "s.nio_bytes_w += len(_r.content) if (_r and _r.status_code < 400) else 0",
                    )
                else:
                    d = d.replace("🐚", "")
                itree.exe(d, locals())
        except Exception as e:
            print(f"🦉 {module_name}.{class_name}.{func_name} is not traced({e})")
            if s.verbose:
                print(traceback.format_exc())
            # sys.exit(0)
        finally:
            s.count += 1

    def empty(s):
        return not s.tau

    def get_metrics(s, metrics_key=KEY_LATENCY) -> List[Tree]:
        r = []
        for k0 in s.tau:
            r.append(s.tau[k0][metrics_key])
        return r

    def get_k0s(s) -> List[str]:
        return s.tau.keys()

    def get_k0s_summary(s, metrics_key=KEY_LATENCY) -> List[Tuple]:
        ks = s.tau.keys()
        r = []
        for k0 in ks:
            t = s.tau[k0][metrics_key]
            t.consolidate()
            r.append((k0, t.root.span(), t.root.start, t.root.end))
        return r

    def get_metrics_by_k0(s, k0=None, metrics_key=KEY_LATENCY) -> Union[Tree, None]:
        return s.tau[k0][metrics_key] if k0 in s.tau else None

    def show(s, ignore_empty_tree=False, show_key=False, time_format=FORMAT_DATETIME):
        for k0 in s.tau:
            for k1 in s.tau[k0]:
                if not isinstance(s.tau[k0][k1], Tree):
                    continue
                s.tau[k0][k1].consolidate()
                span = abs(s.tau[k0][k1].root.span())
                # print(f"{span=:.9f}")
                if not ignore_empty_tree or span > 1e-6:
                    if show_key:
                        print(f"🔑 k0: {k0}, 🗝 k1: {k1}")
                    print(s.tau[k0][k1].get_graph(time_format=time_format))

    def get_overhead(s, format_=OverHeadFormat.ABS) -> float:
        """get tracing latency overhead in absolute format or percentage format

        Args:
            `format_` (str, optional): "abs" - output absolute value in milliseconds; otherwise, output in float point format: `overhead/total_latency`. Defaults to "abs".

        Returns:
            float: absolute value in micro-second, or a number between 0 and 1 which means the percentage of overhead over total latency.
        """
        if format_ == OverHeadFormat.ABS:
            return s.get_overhead_us()
        elif format == OverHeadFormat.PCT:
            return s.get_overhead_pct()
        else:
            raise ValueError(f"wrong format: {format_}")

    def get_overhead_us(s) -> float:
        """get tracing latency overhead in unit of micro-second

        How to calculate latency overhead? The latency overhead is attached to the latency tree. the latency overhead calculation is based on monotonic time and has unit of micro-second. When a HiQ system is instantiated, the initial overhead is 0. Every time a target function is called, we accumulate the overhead. When we finish a level-2 node in the HiQ tree, we update the `overhead` in the tree to get the final overhead of that trace.

        Returns:
            float: absolute value in micro-second
        """
        return s.overhead_us

    def get_overhead_pct(s) -> float:
        """get tracing latency overhead in percentage format

        Returns:
            float: a number between 0 and 1 which means the percentage of overhead over total latency.
        """
        total_latency = 0
        for k0 in s.tau:
            for k1 in s.tau[k0]:
                if k1 == KEY_LATENCY:
                    if itree.is_virtual_node(s.tau[k0][k1].root):
                        itree._itree.consolidate(s.tau[k0][k1].root)
                    tmp = s.tau[k0][k1].root.span()
                    if tmp == tmp:
                        total_latency += tmp
        if total_latency > 0:
            return s.overhead_us / (total_latency * 1e6)
        else:
            return -1

    @abstractmethod
    def custom(self):
        """The abstract method for customizing tracing logic"""

    @abstractmethod
    def custom_disable(self):
        """The abstract method for disabling customized tracing logic"""

    def __getstate__(self):
        self.get_tau_id = None
        a_state = itree.ForestStats.__getstate__(self)
        b_state = self.__dict__
        return (a_state, b_state)

    def __setstate__(self, state):
        a_state, b_state = state
        self.__dict__ = b_state
        itree.ForestStats.__setstate__(self, a_state)


class HiQLatency(HiQBase, Jack):
    """
    A convenient class for latency tracking

    Args:
        hiq_table_or_path (Union[str, List[Iterable[str]]], optional): this is an HiQ Conf, please refer to *HiQ Core Concepts* section in the HiQ documentation.Defaults to [].
        metric_funcs (List[Callable], optional): simple metric function with empty argument. Defaults to [time.time].
        hiq_id_func (Callable, optional): a callable to generate unique id for tau, the hiq map. Defaults to hiq.hiq_utils.get_tau_id.
        func_args_handler(Callable, optional): a callable to convert function args/kwargs into a string. Defaults to hiq.hiq_utils.func_args_handler.
        target_path (str, optional): the directory of the target code. Defaults to None.
        max_hiq_size (int, optional): the max size of hiq map. if the number is exceeded, tree will be sent to LMK. Defaults to 30.
        verbose (bool, optional): when verbose is true, more information will be recorded, like the full stack trace of exception will be recorded in HiQ tree node. Defaults to False.
        fast_fail (bool, optional): when it is true, raise exception to the upper level, don't swallow exceptions. Defaults to True.
        tpl (str, optional): hiq tpl path. Defaults to None.
        extra_hiq_table (List[str], optional): a list of string to decide to include metrics other than latency. Defaults to [].
        attach_timestamp (bool, optional): for non-time/non-latency metric tree, should we attach start, end timestamps in the extra dictionary? Defaults to False.
        extra_metrics (set, optional): metrics to track in `extra` field of HiQ node. The value could be EXTRA_METRIC_ARGS, EXTRA_METRIC_FILE or EXTRA_METRIC_FUNC and it is normally used in development environment. Defaults to set().

    Raises:
        ValueError: Requires the input is valid

    Example usage:

    >>> from hiq.base import HiQLatency
    >>> trace = HiQLatency()
    """

    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time],
        hiq_id_func: Callable = get_tau_id,
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
    ):
        HiQBase.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=hiq_id_func,
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )
        Jack.__init__(sf)

    def custom(sf):
        pass

    def custom_disable(sf):
        pass


class HiQMemory(HiQBase, Jack):
    """
    A convenient class for RSS(resident set size) of memory tracking. Unit of memory is MB.

    RSS is the portion of memory occupied by a process that is held in main memory (RAM). The rest of the occupied memory exists in the swap space or file system, either because some parts of the occupied memory were paged out, or because some parts of the executable were never loaded.

    Args:
        hiq_table_or_path (Union[str, List[Iterable[str]]], optional): this is an HiQ Conf, please refer to *HiQ Core Concepts* section in the HiQ documentation.Defaults to [].
        metric_funcs (List[Callable], optional): simple metric function with empty argument. Defaults to [time.time, get_memory_mb].
        hiq_id_func (Callable, optional): a callable to generate unique id for tau, the hiq map. Defaults to hiq.hiq_utils.get_tau_id.
        func_args_handler(Callable, optional): a callable to convert function args/kwargs into a string. Defaults to hiq.hiq_utils.func_args_handler.
        target_path (str, optional): the directory of the target code. Defaults to None.
        max_hiq_size (int, optional): the max size of hiq map. if the number is exceeded, tree will be sent to LMK. Defaults to 30.
        verbose (bool, optional): when verbose is true, more information will be recorded, like the full stack trace of exception will be recorded in HiQ tree node. Defaults to False.
        fast_fail (bool, optional): when it is true, raise exception to the upper level, don't swallow exceptions. Defaults to True.
        tpl (str, optional): hiq tpl path. Defaults to None.
        extra_hiq_table (List[str], optional): a list of string to decide to include metrics other than latency. Defaults to [].
        attach_timestamp (bool, optional): for non-time/non-latency metric tree, should we attach start, end timestamps in the extra dictionary? Defaults to False.
        extra_metrics (set, optional): metrics to track in `extra` field of HiQ node. The value could be EXTRA_METRIC_ARGS, EXTRA_METRIC_FILE or EXTRA_METRIC_FUNC and it is normally used in development environment. Defaults to set().

    Raises:
        ValueError: Requires the input is valid

    Example usage:

    >>> from hiq.base import HiQMemory
    >>> trace = HiQMemory()
    """

    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time, get_memory_mb],
        hiq_id_func: Callable = get_tau_id,
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),  # EXTRA_METRIC_FILE
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
    ):
        HiQBase.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=hiq_id_func,
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )
        Jack.__init__(sf)

    def custom(sf):
        pass

    def custom_disable(sf):
        pass


HiQSimple = HiQLatency

if __name__ == "__main__":
    import pickle
    from multiprocessing import Process, Queue, Lock

    def get_tau_id_():
        return 1

    h1 = HiQLatency(fast_fail=1, hiq_id_func=get_tau_id_)
    h1.tau[0] = {"hello": Tree(extra={}, tid='simon', queue_lmk=Queue())}
    h1.tau[1] = {"hello": Tree(extra={}, tid='simon', queue_lmk=None)}
    h1.tau[2] = {"hello": 123}

    d = pickle.dumps(h1)
    print(d)
    h2 = pickle.loads(d)
    d2 = pickle.dumps(h2)
    # print(d2)
    print(d == d2)

    # print(h1.tau)
    print(type(h2))
    print(h2.tau)
    print(type(h1.tau[0]['hello']))
    print(type(h2.tau[0]['hello']))
    print(h1.tau[0]['hello'].queue_lmk)
    print(h2.tau[0]['hello'].queue_lmk)
    h2.show()



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import os
import hiq
import itree
from hiq import tree_func
from hiq.constants import FORMAT_DATETIME, FORMAT_TIMESTAMP


itree.uuid()

# ZIN_THRESHOLD will overwrite the default value
global_zin_threshold = None


def init(
    sf,
    s="",
    tid="",
    extra=None,
    monotonic=True,
    capacity=1024,
    queue_lmk=None,
):
    """Constructor of Tree
    Args:
        s (str, optional): a string representation which a tree could be recovered from. Defaults to "".
        tid (str, optional): a unique id of the tree. Defaults to None.
        extra (dict, optional): extra information of the tree. Defaults to {}.
    """
    if not extra:
        extra = {}
    global global_zin_threshold
    if global_zin_threshold is None:
        global_zin_threshold = float(os.environ.get("ZIN_THRESHOLD", 1e-7))

    if not tid:
        # latency: 50us
        tid = itree.uuid()
    # interesting, the old syntax is required here :-)
    itree.Tree.__init__(
        sf,
        tid=tid,
        extra=extra,
        monotonic=monotonic,
        capacity=capacity,
        zin_threshold=global_zin_threshold,
    )

    if s:
        sf.recover(s)
    sf.queue_lmk = queue_lmk


def get_map():
    _map = {"__init__": init, "__slots__": ("queue_lmk",)}
    PREFIX = "tree_"
    for i in dir(tree_func):
        if i.startswith(PREFIX):
            itree.exe(f"_map[i[{len(PREFIX)}:]] = tree_func.{i}", locals())
    return _map


Tree = type(
    "Tree",
    (itree.Tree,),
    get_map(),
)


def get_graph_from_string(
    s: str, time_format=FORMAT_TIMESTAMP, whole_node_name="__predict", with_ordered_node=True
) -> str:
    """deserialize a string back into a HiQ tree and return the graph representation and ordered duration to the caller
    
    
    >>> m = "t1^get_memory_mb,,0,54,5,0,1e-07,0#%n1*[None,inf,inf,0$0#[__predict,9942.6796875,10064.375,0$0#[__pdf,9942.6796875,10013.578125,0$0#][__txt,10013.578125,10175.28515625,0$0#[__det,10013.578125,10174.68359375,0$0#[__ort,10228.8984375,10357.7421875,0$0#]][__det,10174.68359375,10175.28515625,0$0#[__ort,10174.68359375,10175.28515625,0$0#]]][_ort_sess,10175.28515625,10255.44921875,0$0#]]]"
    >>> print(get_graph_from_string(m))
    [      9942.680 -      10064.375]  [100.00%] 🟢_root_get_memory_mb(121.6953)
    [      9942.680 -      10064.375]  [100.00%]    l___predict(121.6953)
    [      9942.680 -      10013.578]  [ 58.26%]       |___pdf(70.8984)
    [     10013.578 -      10175.285]  [132.88%]       |___txt(161.7070)
    [     10013.578 -      10174.684]  [132.38%]       |  |___det(161.1055)
    [     10228.898 -      10357.742]  [105.87%]       |  |  l___ort(128.8438)
    [     10174.684 -      10175.285]  [  0.49%]       |  l___det(0.6016)
    [     10174.684 -      10175.285]  [  0.49%]       |     l___ort(0.6016)
    [     10175.285 -      10255.449]  [ 65.87%]       l__ort_sess(80.1641)
    ordered lowest level calls:__ort:129.4453, _ort_sess:80.1641, __pdf:70.8984

    """
    t = Tree(s)
    entire_latency = t.get_duration_by_node_name(whole_node_name)
    tracking_overhead = float("-inf")  # t.get_overhead()
    if "overhead" in t.extra:
        tracking_overhead = float(t.extra["overhead"]) / (entire_latency * 1e4)
    graph = t.get_graph(time_format)
    res = graph
    if with_ordered_node:
        ordered_duration = t.leaf_duration_by_name_s(ordered=3)
        res += f"ordered lowest level calls:{ordered_duration}\n"
    if "overhead" in t.extra:
        res += f"{tracking_overhead=:.3f}%"
    return res


def get_duration_from_hiq_string(s: str, key: str) -> float:
    """
    >>> m = "t1^get_memory_mb,,0,54,5,0,1e-07,0#%n1*[None,inf,inf,0$0#[__predict,9942.6796875,10064.375,0$0#[__pdf,9942.6796875,10013.578125,0$0#][__txt,10013.578125,10175.28515625,0$0#[__det,10013.578125,10174.68359375,0$0#[__ort,10228.8984375,10357.7421875,0$0#]][__det,10174.68359375,10175.28515625,0$0#[__ort,10174.68359375,10175.28515625,0$0#]]][_ort_sess,10175.28515625,10255.44921875,0$0#]]]"
    >>> print(get_duration_from_hiq_string(m, "__txt"))
    161.70703125
    >>> print(get_duration_from_hiq_string(m, "__ort"))
    129.4453125
    """
    t = Tree(s)
    node_names = t.get_distinct_node_names()
    if key not in node_names:
        raise ValueError(
            f"😬 could not find {key} in hiq tree. hiq_str: {s}, valid nodes: {node_names}"
        )
    return t.get_duration_by_node_name(key)


if __name__ == "__main__":
    m = "t1^get_memory_mb,,0,54,5,0,1e-07,0#%n1*[None,inf,inf,0$0#[__predict,9942.6796875,10064.375,0$0#[__pdf,9942.6796875,10013.578125,0$0#][__txt,10013.578125,10175.28515625,0$0#[__det,10013.578125,10174.68359375,0$0#[__ort,10228.8984375,10357.7421875,0$0#]][__det,10174.68359375,10175.28515625,0$0#[__ort,10174.68359375,10175.28515625,0$0#]]][_ort_sess,10175.28515625,10255.44921875,0$0#]]]"
    print(get_graph_from_string(m, with_ordered_node=False))
    
    print(get_graph_from_string(m, time_format=FORMAT_DATETIME, whole_node_name='__txt'))
    
    print(get_duration_from_hiq_string(m, "__ort"))
    
    import pickle

    a = Tree("", "tid_123", {})

    print(a)
    print(dir(a))
    r = a.get_root()
    print(r)

    d = pickle.dumps(a)
    print(d)
    h2 = pickle.loads(d)
    d2 = pickle.dumps(h2)
    # print(d2)
    print(d == d2)



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
#

import time

import hiq
from hiq import HiQLatency, HiQMemory
from hiq.constants import *
from hiq.hiq_utils import FlaskReqIdGenerator, func_args_handler
from hiq.memory import get_memory_mb

from typing import Union, Callable, List
from hiq.hiq_utils import (
    func_args_handler,
    get_tau_id,
)


class HiQFlaskLatency(HiQLatency):
    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time],
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
    ):
        HiQLatency.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=FlaskReqIdGenerator(),
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )

    def custom(sf):
        pass

    def custom_disable(sf):
        pass


class HiQFlaskMemory(HiQMemory):
    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time, get_memory_mb],
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
    ):
        HiQLatency.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=FlaskReqIdGenerator(),
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )

    def custom(sf):
        pass

    def custom_disable(sf):
        pass


class HiQFlaskLatencyOtel(hiq.HiQSimple):
    def __init__(
        sf,
        hiq_table_or_path: Union[str, list] = [],
        metric_funcs: List[Callable] = [time.time],
        hiq_id_func: Callable = get_tau_id,
        func_args_handler: Callable = func_args_handler,
        target_path=None,
        max_hiq_size=30,
        verbose=False,
        fast_fail=True,
        tpl=None,
        extra_hiq_table=[],
        attach_timestamp=False,
        extra_metrics=set(),
        lmk_path=None,
        lmk_handler=None,
        lmk_logger=None,
        trace=None,
        trace_provider=None,
        tracer_name=None,
        span_processor=None,
        exporter=None,
        endpoints=None,
    ):
        """In addition to arguments of normal HiQLatency, there are several new arguments for OpenTelemetry:
        trace, trace_provider, tracer_name, span_processor, exporter, endpoints
        
        endpoints - server endpoints on which metrics will be sent
        """
        hiq.HiQSimple.__init__(
            sf,
            hiq_table_or_path=hiq_table_or_path,
            metric_funcs=metric_funcs,
            hiq_id_func=hiq_id_func,
            func_args_handler=func_args_handler,
            target_path=target_path,
            max_hiq_size=max_hiq_size,
            verbose=verbose,
            fast_fail=fast_fail,
            tpl=tpl,
            extra_hiq_table=extra_hiq_table,
            attach_timestamp=attach_timestamp,
            extra_metrics=extra_metrics,
            lmk_path=lmk_path,
            lmk_handler=lmk_handler,
            lmk_logger=lmk_logger,
        )
        sf.trace = trace or hiq.mod("opentelemetry.trace")
        sf.trace.set_tracer_provider(
            trace_provider or hiq.mod("opentelemetry.sdk.trace").TracerProvider()
        )
        sf.tracer = sf.trace.get_tracer_provider().get_tracer(tracer_name or __name__)
        sf.exporter = (
            exporter or hiq.mod("opentelemetry.sdk.trace.export").ConsoleSpanExporter()
        )
        sf.span_processor = span_processor or hiq.mod(
            "opentelemetry.sdk.trace.export"
        ).BatchSpanProcessor(sf.exporter)

        sf.trace.get_tracer_provider().add_span_processor(sf.span_processor)
        sf.endpoints = endpoints

    def custom(s):
        s.o_flask_dispatch_request = hiq.mod("flask").Flask.dispatch_request

        def flask_dispatch_request(self):
            from flask import request
            from opentelemetry.instrumentation.wsgi import collect_request_attributes
            from opentelemetry.propagate import extract

            if s.endpoints is None or str(request.endpoint) not in s.endpoints:
                return s.o_flask_dispatch_request(self)

            req_env = collect_request_attributes(request.environ)
            with s.tracer.start_as_current_span(
                str(request.endpoint),
                context=extract(request.headers),
                kind=s.trace.SpanKind.SERVER,
                attributes=req_env,
            ):
                # print(req_env)
                return s.o_flask_dispatch_request(self)

        hiq.mod("flask").Flask.dispatch_request = flask_dispatch_request

    def custom_disable(s):
        hiq.mod("flask").Flask.dispatch_request = s.o_flask_dispatch_request



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import ast
import io
import operator
import os

import sys
import traceback
from collections import defaultdict
from typing import List

import itree
from hiq.constants import FORMAT_TIMESTAMP
from hiq.node_utils import _c, _d
from hiq.tree_utils import split_extra
from hiq.node_utils import __peek_tree, _pp_tree
from hiq.utils import get_env_bool


def get_dict(**kwargs):
    return kwargs


def tree_start(sf, a, b, extra=None):
    if extra is None:
        extra = {}
    sf.discover(a, b, extra)
    if sf.queue_lmk:
        sf.queue_lmk.put_nowait(
            get_dict(id_=sf.tid, name=a, value=b, extra=extra, is_start=True)
        )


def tree_end(sf, a, b, extra=None):
    if extra is None:
        extra = {}
    try:
        sf.finish(a, b, extra)
    except Exception as e:
        traceback.print_exc(file=sys.stdout)
    if sf.queue_lmk:
        sf.queue_lmk.put_nowait(
            get_dict(id_=sf.tid, name=a, value=b, extra=extra, is_start=False)
        )


def tree_consolidate(sf):
    if len(sf.stk) > 1:
        # print("no need to consolidate an incomplete tree")
        return
    sf.root = itree._itree.consolidate(sf.root)


def tree_recover(sf, s):
    if s:
        sf.deserialize(s)


def tree_repr(sf) -> str:
    return itree._itree.serialize_tree_(sf)


def tree_total_duration_s(sf):
    if sf.monotonic:
        return sf.root.span()
    else:
        raise NotImplementedError("TODO")


def tree_duration_by_name_s(sf) -> dict:
    if not sf.root:
        return {}
    visited = [sf.root]
    stack = [sf.root]
    res = defaultdict(float)
    res[sf.root.name] = sf.root.span()
    while stack:
        nd = stack[-1]
        if nd not in visited:
            visited.append(nd)
        remove_from_stack = True
        for nex in nd.nodes:
            if nex not in visited:
                stack.append(nex)
                res[nex.name] += nex.span()
                remove_from_stack = False
                break
        if remove_from_stack:
            stack.pop()
    return dict(res)


def tree_leaf_duration_by_name_s(sf, ordered=0):
    if not sf.root:
        return {}
    visited = [sf.root]
    stack = [sf.root]
    res = defaultdict(float)
    if not sf.root.nodes:
        res[sf.root.name] = sf.root.span()
    while stack:
        nd = stack[-1]
        if nd not in visited:
            visited.append(nd)
        remove_from_stack = True
        for nex in nd.nodes:
            if nex not in visited:
                stack.append(nex)
                remove_from_stack = False
                break
        if remove_from_stack and stack:
            if stack[-1] and not stack[-1].nodes:  # leaf node
                res[stack[-1].name] += stack[-1].span()
            stack.pop()
    res = dict(res)
    if ordered == 1:
        return dict(sorted(res.items(), key=operator.itemgetter(1), reverse=True))
    elif ordered == 2:
        return dict(sorted(res.items(), key=operator.itemgetter(1), reverse=False))
    elif ordered == 3:
        dt = dict(sorted(res.items(), key=operator.itemgetter(1), reverse=True))
        sio = io.StringIO()
        for key, value in dt.items():
            print(f"{key}:{value:6.4f}, ", end="", file=sio)
        return sio.getvalue().strip(", ")
    return dict(res)


def tree_leaf_duration_total_s(sf) -> float:
    res = sf.leaf_duration_by_name_s()
    total = sum(res.values())
    return total


def tree_get_distinct_node_names(sf) -> List[str]:
    r = set()

    def __d(n):
        if not n:
            return
        nonlocal r
        r.add(n.name)
        for i in n.nodes:
            __d(i)

    __d(sf.root)
    r.discard(None)
    return list(r)


def tree_get_duration_by_node_name(s, name) -> float:
    r = 0.0

    def __d(n):
        if not n:
            return
        nonlocal r
        if n.name == name:
            r += n.span()
        [__d(i) for i in n.nodes]

    __d(s.root)
    return r


def tree_get_duration_by_node_name_and_level(s, name, level) -> float:
    """root(None node) is level 0"""
    r = 0.0
    q, level_ = [s.root], 0
    while len(q) > 0:
        tmp = []
        for i in q:
            for j in i.nodes:
                tmp.append(j)
                if level_ == level:
                    if j.name == name:
                        r += j.span()
        level_ += 1
        q = tmp
    return r


def tree_show(s, time_format=FORMAT_TIMESTAMP):
    print(s.get_graph(time_format))


def tree_get_graph(sf, time_format=FORMAT_TIMESTAMP) -> str:
    sf.consolidate()
    sio = io.StringIO()
    wide_output, value_column_len, _type = __peek_tree(sf.root)
    wide_output = get_env_bool("WIDE_OUTPUT", wide_output)
    _pp_tree(
        sf.root,
        time_format=time_format,
        file=sio,
        root_name=sf.tid,
        wide_output=wide_output,
        value_column_len=value_column_len,
        _type=_type,
        tree_extra=sf.extra,
    )
    return sio.getvalue()


def tree_just_before_ending_root(s):
    return len(s.stk) <= 2


# It is pickale, and serializable!!!
# In addition to pickling, for multiprocess code, we can pass the `repr()` between processes.
# There lmk_queue will be lost after pickling :-)
def tree___getstate__(self):
    a_state = itree.Tree.__getstate__(self)
    b_state = self.__dict__
    return (a_state, b_state)


def tree___setstate__(self, state):
    a_state, b_state = state
    self.__dict__ = b_state
    self.queue_lmk = None
    itree.Tree.__setstate__(self, a_state)



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import psutil


def get_memory_mb() -> float:
    """Get RSS in MB unit

    Returns:
        float: RSS of current process in MB
    """
    return psutil.Process().memory_info().rss / (1024 * 1024)


def get_memory_kb() -> float:
    return psutil.Process().memory_info().rss / 1024


def get_memory_b() -> float:
    return psutil.Process().memory_info().rss



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import oci

from base64 import b64encode
from hiq.utils import Singleton
import os


class OciStreamingClient(metaclass=Singleton):
    def __init__(
        self,
        ociMessageEndpoint=None,
        ociStreamOcid=None,
        ociConfigFilePath="~/.oci/config",
        ociProfileName="DEFAULT",
    ):
        if ociMessageEndpoint is None:
            ociMessageEndpoint = os.environ.get("OCI_STM_END", "")
        if ociStreamOcid is None:
            ociStreamOcid = os.environ.get("OCI_STM_OCID", "")
        if not ociMessageEndpoint:
            raise ValueError("🦉 Please setup environment variable `OCI_STM_END`")
        if not ociStreamOcid:
            raise ValueError("🦉 Please setup environment variable `OCI_STM_OCID`")
        self.ociMessageEndpoint = ociMessageEndpoint
        self.ociStreamOcid = ociStreamOcid
        self.ociConfigFilePath = ociConfigFilePath
        self.ociProfileName = ociProfileName

        self.config = oci.config.from_file(self.ociConfigFilePath, self.ociProfileName)
        self.client = oci.streaming.StreamClient(
            self.config, service_endpoint=ociMessageEndpoint
        )

    def __create_message(self, key, message):
        encoded_key = b64encode(key.encode()).decode()
        encoded_value = b64encode(message.encode()).decode()
        return [
            oci.streaming.models.PutMessagesDetailsEntry(
                key=encoded_key, value=encoded_value
            )
        ]

    def produce_messages(self, key, message, check=False):
        """produce message to OCI streaming

        Emits messages to a stream. There's no limit to the number of messages in a request, but the total size of a message or request must be 1 MiB or less. The service calculates the partition ID from the message key and stores messages that share a key on the same partition. If a message does not contain a key or if the key is null, the service generates a message key for you. The partition ID cannot be passed as a parameter.

        Args:
            key (str): message key
            message (str): message value
            check (bool, optional): if True, check the response from OCI streaming server. Defaults to False.
        """
        message_list = self.__create_message(key, message)
        messages = oci.streaming.models.PutMessagesDetails(messages=message_list)
        put_message_result = self.client.put_messages(self.ociStreamOcid, messages)

        if check:
            for entry in put_message_result.data.entries:
                if entry.error:
                    print("Error ({}) : {}".format(entry.error, entry.error_message))
                else:
                    print(
                        "Published message to partition {} , offset {}".format(
                            entry.partition, entry.offset
                        )
                    )


if __name__ == "__main__":
    client = OciStreamingClient()
    import time

    start = time.monotonic()
    client.produce_messages("hiq", "v2main,[🍁 n=fruit,s=0.00,e=40.00,x=0,c=0]")
    end = time.monotonic()
    # 1396091.409958899 super slow
    print((end - start) * 1e6)



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

import io
from hiq.node_utils import __peek_tree
from hiq.node_utils import _pp_tree
from hiq.constants import *
from typing import Tuple


def split_extra(s: str) -> Tuple[str, str]:
    """
    Returns:
        tree_extra, pure tree
    """
    if not s.startswith("v"):
        return None, s
    else:
        s = s[s.index(",") + 1 :]
        run_length = int(s[: s.index(",")])
        s = s[s.index(",") + 1 :]
        return s[:run_length], s[run_length:]



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

from prometheus_client import Summary, push_to_gateway, CollectorRegistry, Counter
from hiq.utils import memoize_first
import asyncio


@memoize_first
def get_summary(name, desc=None):
    if name[0:2] == "__":
        name = "hiq_" + name[2:]
    return Summary(name, name if desc is None else desc)


"""
class GatewayClient:
    #https://stackoverflow.com/questions/61785170/how-can-i-expose-prometheus-metrics-from-multiple-subprocesses
    def __init__(self, addr, thing_id):
        self.gateway_address = addr
        self.job_id = "thing-{}".format(thing_id)
        self.registry = CollectorRegistry()
        # define metrics etc
        self.some_counter = Counter('some_counter', 'A counter', registry=self.registry)

    async def start(self):
        while True:
            # push latest data
            push_to_gateway(self.gateway_address, job=self.job_id, registry=self.registry
            await asyncio.sleep(15)

    # methods to set metrics etc
    def inc_some_counter(self):
        self.some_counter.inc()
"""



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

"""A generic node for HiQ tree

Node in HiQ has `start` and `end` property. They can be anything that happens at start/end time,
including time itself, and CPU memory usage, GPU memory usage, io traffic by bytes etc.

extra is a dictionary to record extra information.
"""

import itree

Node = itree.Node
is_virtual_node = itree.is_virtual_node

if __name__ == "__main__":
    n = Node("", 0, 0)
    print(n)



import hiq


class PaddleHiQLatency(hiq.HiQSimple):
    def custom(s):
        s.o_paddle_run = hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run

        @s.inserter
        def paddle_run(self) -> bool:
            return s.o_paddle_run(self)

        hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run = paddle_run

    def custom_disable(s):
        hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run = s.o_paddle_run


class PaddleHiQMemory(hiq.HiQMemory):
    def custom(s):
        s.o_paddle_run = hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run

        @s.inserter
        def paddle_run(self) -> bool:
            return s.o_paddle_run(self)

        hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run = paddle_run

    def custom_disable(s):
        hiq.mod('paddle.fluid.core_avx').PaddleInferPredictor.run = s.o_paddle_run



# HiQ version 1.1.2
#
# Copyright (c) 2022, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/ 
#

__author__ = "Fuheng Wu<fuheng.wu@oralce.com>"
__date__ = "2023-01-01"
__doc__ = "HiQ is a declarative, non-intrusive, dynamic and transparent tracking and optimization system"
__version__ = "1.1.2"
__credits__ = "Henry Wu, Ivan Davchev, Jun Qian"

import sys

vinfo = sys.version_info
if vinfo.major < 3 or (vinfo.major == 3 and vinfo.minor < 8):
    print(f"HiQ supports Python version 3.8 and higher. Your version is {vinfo.major}.{vinfo.minor}.")
    sys.exit(1)

from .hiq_utils import (
    set_global_hiq_status,
    get_global_hiq_status,
    HiQStatusContext,
)
from .utils import (
    SingletonMeta,
    SingletonBase,
    memoize,
    memoize_first,
    lmk_data_handler,
    is_hiqed,
    get_env_bool,
    get_env_int,
    get_env_float,
    get_proxies,
    SilencePrint,
    random_str,
    download_from_http,
    ensure_folder,
    write_file,
    read_file,
    ts_pair_to_dt,
    ts_to_dt,
    utc_to_pst,
    execute_cmd,
    get_home,
    create_gantt_chart_time,
    create_gantt_chart_memory
)
from .tree import (
    get_duration_from_hiq_string,
    get_graph_from_string,
    Tree
)

try:
    from .memory import (
        get_memory_mb,
        get_memory_kb,
        get_memory_b,
    )
except ImportError:
    pass
from .base import HiQLatency, HiQSimple, HiQMemory
from .server_flask import HiQFlaskLatency, HiQFlaskMemory, HiQFlaskLatencyOtel
from .server_fastapi import HiQFastAPILatency, HiQFastAPILatencyMixin, HiQFastAPIMemory, run_fastapi
from .constants import *
from .ddict import *

try:
    from itree import mod
except ImportError:
    pass

__all__ = [
    "HiQLatency",
    "HiQFlaskLatency",
    "HiQFlaskMemory",
    "HiQFlaskLatencyOtel",
    "HiQFastAPILatency",
    "HiQFastAPIMemory",
    "HiQFastAPILatencyMixin",
    "run_fastapi",
    "HiQSimple",
    "HiQMemory",
    "HiQStatusContext",
    "get_global_hiq_status",
    "set_global_hiq_status",
    "SingletonMeta",
    "SingletonBase",
    "memoize",
    "memoize_first",
    "lmk_data_handler",
    "is_hiqed",
    "get_env_bool",
    "get_env_int",
    "get_env_float",
    "get_proxies",
    "SilencePrint",
    "random_str",
    "download_from_http",
    "ensure_folder",
    "write_file",
    "read_file",
    "get_memory_mb",
    "get_memory_kb",
    "get_memory_b",
    "ts_pair_to_dt",
    "ts_to_dt",
    "utc_to_pst",
    "execute_cmd",
    "get_home",
    "get_duration_from_hiq_string",
    "get_graph_from_string",
    "Tree",
    "create_gantt_chart_time",
    "create_gantt_chart_memory"
]
